# Copyright (c) 2022 NetApp, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from datetime import datetime
from datetime import timedelta
import math
from time import time

from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import units
import six

from cinder import exception
from cinder.i18n import _
from cinder import utils
from cinder.volume.drivers.netapp.dataontap.client import api as netapp_api
from cinder.volume.drivers.netapp.dataontap.client import client_cmode
from cinder.volume.drivers.netapp import utils as na_utils
from cinder.volume import volume_utils

LOG = logging.getLogger(__name__)
DEFAULT_MAX_PAGE_LENGTH = 10000
ONTAP_SELECT_MODEL = 'FDvM300'
ONTAP_C190 = 'C190'
HTTP_ACCEPTED = 202
DELETED_PREFIX = 'deleted_cinder_'
DEFAULT_TIMEOUT = 15
REST_SYNC_TIMEOUT = 15

# Keys in this map are REST API's endpoints that the user shall have permission
# in order to enable extra specs reported to Cinder's scheduler.
# NOTE(sfernand): ONTAP does not retrieve volume efficiency information
# properly when using the pre-created "vsadmin" role (SVM scoped), causing
# dedup and compression extra specs to be reported as disabled despite its
# current configuration.
SSC_API_MAP = {
    '/storage/aggregates': [
        'netapp_raid_type',
    ],
    '/storage/disks': [
        'netapp_disk_type',
    ],
    '/snapmirror/relationships': [
        'netapp_mirrored',
    ],
    '/storage/volumes': [
        'netapp_flexvol_encryption'
        'netapp_dedup',
        'netapp_compression',
    ],
}


@six.add_metaclass(volume_utils.TraceWrapperMetaclass)
class RestClient(object):

    def __init__(self, **kwargs):

        host = kwargs['hostname']
        username = kwargs['username']
        password = kwargs['password']
        api_trace_pattern = kwargs['api_trace_pattern']
        self.connection = netapp_api.RestNaServer(
            host=host,
            transport_type=kwargs['transport_type'],
            ssl_cert_path=kwargs.pop('ssl_cert_path'),
            port=kwargs['port'],
            username=username,
            password=password,
            api_trace_pattern=api_trace_pattern)

        self.async_rest_timeout = kwargs.get('async_rest_timeout', 60)

        self.vserver = kwargs.get('vserver')
        self.connection.set_vserver(self.vserver)

        ontap_version = self.get_ontap_version(cached=False)
        if ontap_version < (9, 11, 1):
            msg = _('REST Client can be used only with ONTAP 9.11.1 or upper.')
            raise na_utils.NetAppDriverException(msg)
        self.connection.set_ontap_version(ontap_version)

        self.ssh_client = self._init_ssh_client(host, username, password)

        # NOTE(nahimsouza): ZAPI Client is needed to implement the fallback
        # when a REST method is not supported.
        self.zapi_client = client_cmode.Client(**kwargs)

        self._init_features()

    def _init_ssh_client(self, host, username, password):
        return netapp_api.SSHUtil(
            host=host,
            username=username,
            password=password)

    def _init_features(self):
        self.features = na_utils.Features()

        generation, major, minor = self.get_ontap_version()
        ontap_version = (generation, major)

        ontap_9_0 = ontap_version >= (9, 0)
        ontap_9_4 = ontap_version >= (9, 4)
        ontap_9_5 = ontap_version >= (9, 5)
        ontap_9_6 = ontap_version >= (9, 6)
        ontap_9_8 = ontap_version >= (9, 8)
        ontap_9_9 = ontap_version >= (9, 9)

        nodes_info = self._get_cluster_nodes_info()
        for node in nodes_info:
            qos_min_block = False
            qos_min_nfs = False
            if node['model'] == ONTAP_SELECT_MODEL:
                qos_min_block = node['is_all_flash_select'] and ontap_9_6
                qos_min_nfs = qos_min_block
            elif ONTAP_C190 in node['model']:
                qos_min_block = node['is_all_flash'] and ontap_9_6
                qos_min_nfs = qos_min_block
            else:
                qos_min_block = node['is_all_flash'] and ontap_9_0
                qos_min_nfs = node['is_all_flash'] and ontap_9_0

            qos_name = na_utils.qos_min_feature_name(True, node['name'])
            self.features.add_feature(qos_name, supported=qos_min_nfs)
            qos_name = na_utils.qos_min_feature_name(False, node['name'])
            self.features.add_feature(qos_name, supported=qos_min_block)

        self.features.add_feature('SNAPMIRROR_V2', supported=ontap_9_0)
        self.features.add_feature('USER_CAPABILITY_LIST',
                                  supported=ontap_9_0)
        self.features.add_feature('SYSTEM_METRICS', supported=ontap_9_0)
        self.features.add_feature('CLONE_SPLIT_STATUS', supported=ontap_9_0)
        self.features.add_feature('FAST_CLONE_DELETE', supported=ontap_9_0)
        self.features.add_feature('SYSTEM_CONSTITUENT_METRICS',
                                  supported=ontap_9_0)
        self.features.add_feature('ADVANCED_DISK_PARTITIONING',
                                  supported=ontap_9_0)
        self.features.add_feature('BACKUP_CLONE_PARAM', supported=ontap_9_0)
        self.features.add_feature('CLUSTER_PEER_POLICY', supported=ontap_9_0)
        self.features.add_feature('FLEXVOL_ENCRYPTION', supported=ontap_9_0)
        self.features.add_feature('FLEXGROUP', supported=ontap_9_8)
        self.features.add_feature('FLEXGROUP_CLONE_FILE',
                                  supported=ontap_9_9)

        self.features.add_feature('ADAPTIVE_QOS', supported=ontap_9_4)
        self.features.add_feature('ADAPTIVE_QOS_BLOCK_SIZE',
                                  supported=ontap_9_5)
        self.features.add_feature('ADAPTIVE_QOS_EXPECTED_IOPS_ALLOCATION',
                                  supported=ontap_9_5)

        LOG.info('ONTAP Version: %(generation)s.%(major)s.%(minor)s',
                 {'generation': ontap_version[0], 'major': ontap_version[1],
                  'minor': minor})

    def __getattr__(self, name):
        """If method is not implemented for REST, try to call the ZAPI."""
        LOG.debug("The %s call is not supported for REST, falling back to "
                  "ZAPI.", name)
        # Don't use self.zapi_client to avoid reentrant call to __getattr__()
        zapi_client = object.__getattribute__(self, 'zapi_client')
        return getattr(zapi_client, name)

    def _wait_job_result(self, job_url):
        """Waits for a job to finish."""

        interval = 2
        retries = (self.async_rest_timeout / interval)

        @utils.retry(netapp_api.NaRetryableError, interval=interval,
                     retries=retries, backoff_rate=1)
        def _waiter():
            response = self.send_request(job_url, 'get',
                                         enable_tunneling=False)

            job_state = response.get('state')
            if job_state == 'success':
                return response
            elif job_state == 'failure':
                message = response['error']['message']
                code = response['error']['code']
                raise netapp_api.NaApiError(message=message, code=code)

            msg_args = {'job': job_url, 'state': job_state}
            LOG.debug("Job %(job)s has not finished: %(state)s", msg_args)
            raise netapp_api.NaRetryableError(message='Job is running.')

        try:
            return _waiter()
        except netapp_api.NaRetryableError:
            msg = _("Job %s did not reach the expected state. Retries "
                    "exhausted. Aborting.") % job_url
            raise na_utils.NetAppDriverException(msg)

    def send_request(self, action_url, method, body=None, query=None,
                     enable_tunneling=True,
                     max_page_length=DEFAULT_MAX_PAGE_LENGTH,
                     wait_on_accepted=True):

        """Sends REST request to ONTAP.

        :param action_url: action URL for the request
        :param method: HTTP method for the request ('get', 'post', 'put',
            'delete' or 'patch')
        :param body: dict of arguments to be passed as request body
        :param query: dict of arguments to be passed as query string
        :param enable_tunneling: enable tunneling to the ONTAP host
        :param max_page_length: size of the page during pagination
        :param wait_on_accepted: if True, wait until the job finishes when
            HTTP code 202 (Accepted) is returned

        :returns: parsed REST response
        """

        response = None

        if method == 'get':
            response = self.get_records(
                action_url, query, enable_tunneling, max_page_length)
        else:
            code, response = self.connection.invoke_successfully(
                action_url, method, body=body, query=query,
                enable_tunneling=enable_tunneling)

            if code == HTTP_ACCEPTED and wait_on_accepted:
                # get job URL and discard '/api'
                job_url = response['job']['_links']['self']['href'][4:]
                response = self._wait_job_result(job_url)

        return response

    def get_records(self, action_url, query=None, enable_tunneling=True,
                    max_page_length=DEFAULT_MAX_PAGE_LENGTH):
        """Retrieves ONTAP resources using pagination REST request.

        :param action_url: action URL for the request
        :param query: dict of arguments to be passed as query string
        :param enable_tunneling: enable tunneling to the ONTAP host
        :param max_page_length: size of the page during pagination

        :returns: dict containing records and num_records
        """

        # Initialize query variable if it is None
        query = query if query else {}
        query['max_records'] = max_page_length

        _, response = self.connection.invoke_successfully(
            action_url, 'get', query=query,
            enable_tunneling=enable_tunneling)

        # NOTE(nahimsouza): if all records are returned in the first call,
        # 'next_url' will be None.
        next_url = response.get('_links', {}).get('next', {}).get('href')
        next_url = next_url[4:] if next_url else None  # discard '/api'

        # Get remaining pages, saving data into first page
        while next_url:
            # NOTE(nahimsouza): clean the 'query', because the parameters are
            # already included in 'next_url'.
            _, next_response = self.connection.invoke_successfully(
                next_url, 'get', query=None,
                enable_tunneling=enable_tunneling)

            response['num_records'] += next_response.get('num_records', 0)
            response['records'].extend(next_response.get('records'))

            next_url = (
                next_response.get('_links', {}).get('next', {}).get('href'))
            next_url = next_url[4:] if next_url else None  # discard '/api'

        return response

    def get_ontap_version(self, cached=True):
        """Gets the ONTAP version as tuple."""

        if cached:
            return self.connection.get_ontap_version()

        query = {
            'fields': 'version'
        }

        response = self.send_request('/cluster/', 'get', query=query)

        version = (response['version']['generation'],
                   response['version']['major'],
                   response['version']['minor'])

        return version

    def check_api_permissions(self):
        """Check which APIs that support SSC functionality are available."""

        inaccessible_apis = []
        invalid_extra_specs = []

        for api, extra_specs in SSC_API_MAP.items():
            if not self.check_cluster_api(api):
                inaccessible_apis.append(api)
                invalid_extra_specs.extend(extra_specs)

        if inaccessible_apis:
            if '/storage/volumes' in inaccessible_apis:
                msg = _('User not permitted to query Data ONTAP volumes.')
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                LOG.warning('The configured user account does not have '
                            'sufficient privileges to use all needed '
                            'APIs. The following extra specs will fail '
                            'or be ignored: %s.', invalid_extra_specs)

        return invalid_extra_specs

    def check_cluster_api(self, api):
        """Checks the availability of a cluster API.

        Returns True if the specified cluster API exists and may be called by
        the current user.
        """
        try:
            # No need to return any records here since we just want to know if
            # the user is allowed to make the request. A "Permission Denied"
            # error code is expected in case user does not have the necessary
            # permissions.
            self.send_request('%s?return_records=false' % api, 'get',
                              enable_tunneling=False)
        except netapp_api.NaApiError as ex:
            # NOTE(nahimsouza): This function only returns false in case user
            # is not authorized. If other error is returned, it must be
            # handled in the function call that uses the same endpoint.
            if ex.code == netapp_api.REST_UNAUTHORIZED:
                return False

        return True

    def _get_cluster_nodes_info(self):
        """Return a list of models of the nodes in the cluster."""
        query_args = {'fields': 'model,'
                                'name,'
                                'is_all_flash_optimized,'
                                'is_all_flash_select_optimized'}

        nodes = []
        try:
            result = self.send_request('/cluster/nodes', 'get',
                                       query=query_args,
                                       enable_tunneling=False)

            for record in result['records']:
                node = {
                    'model': record['model'],
                    'name': record['name'],
                    'is_all_flash':
                        record['is_all_flash_optimized'],
                    'is_all_flash_select':
                        record['is_all_flash_select_optimized']
                }
                nodes.append(node)
        except netapp_api.NaApiError as e:
            if e.code == netapp_api.REST_UNAUTHORIZED:
                LOG.debug('Cluster nodes can only be collected with '
                          'cluster scoped credentials.')
            else:
                LOG.exception('Failed to get the cluster nodes.')

        return nodes

    def list_flexvols(self):
        """Returns the names of the flexvols on the controller."""

        query = {
            'type': 'rw',
            'style': 'flex*',  # Match both 'flexvol' and 'flexgroup'
            'is_svm_root': 'false',
            'error_state.is_inconsistent': 'false',
            'state': 'online',
            'fields': 'name'
        }

        response = self.send_request(
            '/storage/volumes/', 'get', query=query)

        records = response.get('records', [])
        volumes = [volume['name'] for volume in records]

        return volumes

    def _get_unique_volume(self, records):
        """Get the unique FlexVol or FlexGroup volume from a volume list."""
        if len(records) != 1:
            msg = _('Could not find unique volume. Volumes found: %(vol)s.')
            msg_args = {'vol': records}
            raise exception.VolumeBackendAPIException(data=msg % msg_args)

        return records[0]

    def _get_volume_by_args(self, vol_name=None, vol_path=None,
                            vserver=None, fields=None):
        """Get info from a single volume according to the args."""

        query = {
            'type': 'rw',
            'style': 'flex*',  # Match both 'flexvol' and 'flexgroup'
            'is_svm_root': 'false',
            'error_state.is_inconsistent': 'false',
            'state': 'online',
            'fields': 'name,style'
        }

        if vol_name:
            query['name'] = vol_name
        if vol_path:
            query['nas.path'] = vol_path
        if vserver:
            query['svm.name'] = vserver
        if fields:
            query['fields'] = fields

        volumes_response = self.send_request(
            '/storage/volumes/', 'get', query=query)

        records = volumes_response.get('records', [])
        volume = self._get_unique_volume(records)
        return volume

    def get_flexvol(self, flexvol_path=None, flexvol_name=None):
        """Get flexvol attributes needed for the storage service catalog."""

        fields = ('aggregates.name,name,svm.name,nas.path,'
                  'type,guarantee.honored,guarantee.type,'
                  'space.snapshot.reserve_percent,space.size,'
                  'qos.policy.name,snapshot_policy,language,style')
        unique_volume = self._get_volume_by_args(
            vol_name=flexvol_name, vol_path=flexvol_path, fields=fields)

        aggregate = None
        if unique_volume['style'] == 'flexvol':
            # flexvol has only 1 aggregate
            aggregate = unique_volume['aggregates'][0]['name']
        else:
            aggregate = [aggr["name"]
                         for aggr in unique_volume.get('aggregates', [])]

        qos_policy_group = (
            unique_volume.get('qos', {}).get('policy', {}).get('name'))

        volume = {
            'name': unique_volume['name'],
            'vserver': unique_volume['svm']['name'],
            'junction-path': unique_volume.get('nas', {}).get('path'),
            'aggregate': aggregate,
            'type': unique_volume['type'],
            'space-guarantee-enabled': unique_volume['guarantee']['honored'],
            'space-guarantee': unique_volume['guarantee']['type'],
            'percentage-snapshot-reserve':
                str(unique_volume['space']['snapshot']['reserve_percent']),
            'size': str(unique_volume['space']['size']),
            'qos-policy-group': qos_policy_group,
            'snapshot-policy': unique_volume['snapshot_policy']['name'],
            'language': unique_volume['language'],
            'style-extended': unique_volume['style'],
        }

        return volume

    def is_flexvol_mirrored(self, flexvol_name, vserver_name):
        """Check if flexvol is a SnapMirror source."""

        query = {
            'source.path': vserver_name + ':' + flexvol_name,
            'state': 'snapmirrored',
            'return_records': 'false',
        }

        try:
            response = self.send_request('/snapmirror/relationships/',
                                         'get', query=query)
            return response['num_records'] > 0
        except netapp_api.NaApiError:
            LOG.exception('Failed to get SnapMirror info for volume %s.',
                          flexvol_name)

        return False

    def is_flexvol_encrypted(self, flexvol_name, vserver_name):
        """Check if a flexvol is encrypted."""

        if not self.features.FLEXVOL_ENCRYPTION:
            return False

        query = {
            'encryption.enabled': 'true',
            'name': flexvol_name,
            'svm.name': vserver_name,
            'return_records': 'false',
        }

        try:
            response = self.send_request(
                '/storage/volumes/', 'get', query=query)
            return response['num_records'] > 0
        except netapp_api.NaApiError:
            LOG.exception('Failed to get Encryption info for volume %s.',
                          flexvol_name)

        return False

    def get_aggregate_disk_types(self, aggregate_name):
        """Get the disk type(s) of an aggregate."""
        disk_types = self._get_aggregate_disk_types(aggregate_name)
        return list(disk_types) if disk_types else None

    def _get_aggregate_disk_types(self, aggregate_name):
        """Get the disk type(s) of an aggregate"""

        disk_types = set()

        query = {
            'aggregates.name': aggregate_name,
            'fields': 'effective_type'
        }

        try:
            response = self.send_request(
                '/storage/disks', 'get', query=query, enable_tunneling=False)
        except netapp_api.NaApiError:
            LOG.exception('Failed to get disk info for aggregate %s.',
                          aggregate_name)
            return disk_types

        for storage_disk_info in response['records']:
            disk_types.add(storage_disk_info['effective_type'])

        return disk_types

    def _get_aggregates(self, aggregate_names=None, fields=None):

        query = {}
        if aggregate_names:
            query['name'] = ','.join(aggregate_names)

        if fields:
            query['fields'] = fields

        response = self.send_request(
            '/storage/aggregates', 'get', query=query, enable_tunneling=False)

        return response['records']

    def get_aggregate(self, aggregate_name):
        """Get aggregate attributes needed for the storage service catalog."""

        if not aggregate_name:
            return {}

        fields = ('name,block_storage.primary.raid_type,'
                  'block_storage.storage_type,home_node.name')

        try:
            aggrs = self._get_aggregates(aggregate_names=[aggregate_name],
                                         fields=fields)
        except netapp_api.NaApiError:
            LOG.exception('Failed to get info for aggregate %s.',
                          aggregate_name)
            return {}

        if len(aggrs) < 1:
            return {}

        aggr_attributes = aggrs[0]

        aggregate = {
            'name': aggr_attributes['name'],
            'raid-type':
                aggr_attributes['block_storage']['primary']['raid_type'],
            'is-hybrid':
                aggr_attributes['block_storage']['storage_type'] == 'hybrid',
            'node-name': aggr_attributes['home_node']['name'],
        }

        return aggregate

    def is_qos_min_supported(self, is_nfs, node_name):
        """Check if the node supports QoS minimum."""
        if node_name is None:
            # whether no access to node name (SVM account or error), the QoS
            # min support is dropped.
            return False

        qos_min_name = na_utils.qos_min_feature_name(is_nfs, node_name)
        return getattr(self.features, qos_min_name, False).__bool__()

    def get_flexvol_dedupe_info(self, flexvol_name):
        """Get dedupe attributes needed for the storage service catalog."""

        query = {
            'efficiency.volume_path': '/vol/%s' % flexvol_name,
            'fields': 'efficiency.state,efficiency.compression'
        }

        # Set default values for the case there is no response.
        no_dedupe_response = {
            'compression': False,
            'dedupe': False,
            'logical-data-size': 0,
            'logical-data-limit': 1,
        }

        try:
            response = self.send_request('/storage/volumes',
                                         'get', query=query)
        except netapp_api.NaApiError:
            LOG.exception('Failed to get dedupe info for volume %s.',
                          flexvol_name)
            return no_dedupe_response

        if response["num_records"] != 1:
            return no_dedupe_response

        state = response["records"][0]["efficiency"]["state"]
        compression = response["records"][0]["efficiency"]["compression"]

        # TODO(nahimsouza): as soon as REST API supports the fields
        # 'logical-data-size and 'logical-data-limit', we should include
        # them in the query and set them correctly.
        # NOTE(nahimsouza): these fields are only used by the client function
        # `get_flexvol_dedupe_used_percent`, since the function is not
        # implemented on REST yet, the below hard-coded fields are not
        # affecting the driver in anyway.
        logical_data_size = 0
        logical_data_limit = 1

        dedupe_info = {
            'compression': False if compression == "none" else True,
            'dedupe': False if state == "disabled" else True,
            'logical-data-size': logical_data_size,
            'logical-data-limit': logical_data_limit,
        }

        return dedupe_info

    def get_lun_list(self):
        """Gets the list of LUNs on filer.

        Gets the LUNs from cluster with vserver.
        """

        query = {
            'svm.name': self.vserver,
            'fields': 'svm.name,location.volume.name,space.size,'
                      'location.qtree.name,name,os_type,'
                      'space.guarantee.requested,uuid'
        }

        response = self.send_request(
            '/storage/luns/', 'get', query=query)

        if response['num_records'] == '0':
            return []

        lun_list = []
        for lun in response['records']:
            lun_info = {}
            lun_info['Vserver'] = lun['svm']['name']
            lun_info['Volume'] = lun['location']['volume']['name']
            lun_info['Size'] = lun['space']['size']
            lun_info['Qtree'] = \
                lun['location'].get('qtree', {}).get('name', '')
            lun_info['Path'] = lun['name']
            lun_info['OsType'] = lun['os_type']
            lun_info['SpaceReserved'] = lun['space']['guarantee']['requested']
            lun_info['UUID'] = lun['uuid']

            lun_list.append(lun_info)

        return lun_list

    def get_lun_by_args(self, **lun_info_args):
        """Retrieves LUN with specified args."""

        query = {
            'fields': 'svm.name,location.volume.name,space.size,'
                      'location.qtree.name,name,os_type,'
                      'space.guarantee.requested,uuid'
        }

        if lun_info_args:
            if 'vserver' in lun_info_args:
                query['svm.name'] = lun_info_args['vserver']
            if 'path' in lun_info_args:
                query['name'] = lun_info_args['path']
            if 'uuid' in lun_info_args:
                query['uuid'] = lun_info_args['uuid']

        response = self.send_request(
            '/storage/luns/', 'get', query=query)

        if response['num_records'] == '0':
            return []

        lun_list = []
        for lun in response['records']:
            lun_info = {}
            lun_info['Vserver'] = lun['svm']['name']
            lun_info['Volume'] = lun['location']['volume']['name']
            lun_info['Size'] = lun['space']['size']
            lun_info['Qtree'] = \
                lun['location'].get('qtree', {}).get('name', '')
            lun_info['Path'] = lun['name']
            lun_info['OsType'] = lun['os_type']
            lun_info['SpaceReserved'] = lun['space']['guarantee']['requested']
            lun_info['UUID'] = lun['uuid']

            # NOTE(nahimsouza): Currently, ONTAP REST API does not have the
            # 'block-size' in the response. By default, we are setting its
            # value to 512, since traditional block size advertised by hard
            # disks is 512 bytes.
            lun_info['BlockSize'] = 512

            lun_list.append(lun_info)

        return lun_list

    def get_lun_sizes_by_volume(self, volume_name):
        """"Gets the list of LUNs and their sizes from a given volume name"""

        query = {
            'location.volume.name': volume_name,
            'fields': 'space.size,name'
        }

        response = self.send_request('/storage/luns/', 'get', query=query)

        if response['num_records'] == '0':
            return []

        luns = []
        for lun_info in response['records']:
            luns.append({
                'path': lun_info.get('name', ''),
                'size': float(lun_info.get('space', {}).get('size', 0))
            })
        return luns

    def get_file_sizes_by_dir(self, dir_path):
        """Gets the list of files and their sizes from a given directory."""

        # 'dir_path' will always be a FlexVol name
        volume = self._get_volume_by_args(vol_name=dir_path)

        query = {
            'type': 'file',
            'fields': 'size,name'
        }

        vol_uuid = volume['uuid']
        try:
            response = self.send_request(
                f'/storage/volumes/{vol_uuid}/files',
                'get', query=query)
        except netapp_api.NaApiError as e:
            if e.code == netapp_api.REST_NO_SUCH_FILE:
                return []
            else:
                raise e

        files = []
        for file_info in response['records']:
            files.append({
                'name': file_info.get('name', ''),
                'file-size': float(file_info.get('size', 0))
            })
        return files

    def get_volume_state(self, junction_path=None, name=None):
        """Returns volume state for a given name or junction path."""

        query_args = {}

        if name:
            query_args['name'] = name
        if junction_path:
            query_args['nas.path'] = junction_path

        query_args['fields'] = 'state'

        response = self.send_request('/storage/volumes/',
                                     'get', query=query_args)
        try:
            records = response.get('records', [])
            unique_volume = self._get_unique_volume(records)
        except exception.VolumeBackendAPIException:
            return None

        return unique_volume['state']

    def delete_snapshot(self, volume_name, snapshot_name):
        """Deletes a volume snapshot."""
        volume = self._get_volume_by_args(vol_name=volume_name)
        self.send_request(
            f'/storage/volumes/{volume["uuid"]}/snapshots'
            f'?name={snapshot_name}', 'delete')

    def get_operational_lif_addresses(self):
        """Gets the IP addresses of operational LIFs on the vserver."""

        query = {
            'state': 'up',
            'fields': 'ip.address',
        }

        response = self.send_request(
            '/network/ip/interfaces/', 'get', query=query)

        return [lif_info['ip']['address']
                for lif_info in response['records']]

    def _list_vservers(self):
        """Get the names of vservers present"""
        query = {
            'fields': 'name',
        }
        response = self.send_request('/svm/svms', 'get', query=query,
                                     enable_tunneling=False)

        return [svm['name'] for svm in response.get('records', [])]

    def _get_ems_log_destination_vserver(self):
        """Returns the best vserver destination for EMS messages."""

        # NOTE(nahimsouza): Differently from ZAPI, only 'data' SVMs can be
        # managed by the SVM REST APIs - that's why the vserver type is not
        # specified.
        vservers = self._list_vservers()

        if vservers:
            return vservers[0]

        raise exception.NotFound("No Vserver found to receive EMS messages.")

    def send_ems_log_message(self, message_dict):
        """Sends a message to the Data ONTAP EMS log."""

        body = {
            'computer_name': message_dict['computer-name'],
            'event_source': message_dict['event-source'],
            'app_version': message_dict['app-version'],
            'category': message_dict['category'],
            'severity': 'notice',
            'autosupport_required': message_dict['auto-support'] == 'true',
            'event_id': message_dict['event-id'],
            'event_description': message_dict['event-description'],
        }

        bkp_connection = copy.copy(self.connection)
        bkp_timeout = self.connection.get_timeout()
        bkp_vserver = self.vserver

        self.connection.set_timeout(25)
        try:
            # TODO(nahimsouza): Vserver is being set to replicate the ZAPI
            # behavior, but need to check if this could be removed in REST API
            self.connection.set_vserver(
                self._get_ems_log_destination_vserver())
            self.send_request('/support/ems/application-logs',
                              'post', body=body)
            LOG.debug('EMS executed successfully.')
        except netapp_api.NaApiError as e:
            LOG.warning('Failed to invoke EMS. %s', e)
        finally:
            # Restores the data
            timeout = (
                bkp_timeout if bkp_timeout is not None else DEFAULT_TIMEOUT)
            self.connection.set_timeout(timeout)
            self.connection = copy.copy(bkp_connection)
            self.connection.set_vserver(bkp_vserver)

    def get_performance_counter_info(self, object_name, counter_name):
        """Gets info about one or more Data ONTAP performance counters."""

        # NOTE(nahimsouza): This conversion is nedeed because different names
        # are used in ZAPI and we want to avoid changes in the driver for now.
        rest_counter_names = {
            'domain_busy': 'domain_busy_percent',
            'processor_elapsed_time': 'elapsed_time',
            'avg_processor_busy': 'average_processor_busy_percent',
        }

        rest_counter_name = counter_name
        if counter_name in rest_counter_names:
            rest_counter_name = rest_counter_names[counter_name]

        # Get counter table info
        query = {
            'counter_schemas.name': rest_counter_name,
            'fields': 'counter_schemas.*'
        }

        try:
            table = self.send_request(
                f'/cluster/counter/tables/{object_name}',
                'get', query=query, enable_tunneling=False)

            name = counter_name  # use the original name (ZAPI compatible)
            base_counter = table['counter_schemas'][0]['denominator']['name']

            query = {
                'counters.name': rest_counter_name,
                'fields': 'counters.*'
            }

            response = self.send_request(
                f'/cluster/counter/tables/{object_name}/rows',
                'get', query=query, enable_tunneling=False)

            table_rows = response.get('records', [])
            labels = []
            if len(table_rows) != 0:
                labels = table_rows[0]['counters'][0].get('labels', [])

                # NOTE(nahimsouza): Values have a different format on REST API
                # and we want to keep compatibility with ZAPI for a while
                if object_name == 'wafl' and counter_name == 'cp_phase_times':
                    # discard the prefix 'cp_'
                    labels = [label[3:] for label in labels]

            return {
                'name': name,
                'labels': labels,
                'base-counter': base_counter,
            }
        except netapp_api.NaApiError:
            raise exception.NotFound(_('Counter %s not found') % counter_name)

    def get_performance_instance_uuids(self, object_name, node_name):
        """Get UUIDs of performance instances for a cluster node."""

        query = {
            'id': node_name + ':*',
        }

        response = self.send_request(
            f'/cluster/counter/tables/{object_name}/rows',
            'get', query=query, enable_tunneling=False)

        records = response.get('records', [])

        uuids = []
        for record in records:
            uuids.append(record['id'])

        return uuids

    def get_performance_counters(self, object_name, instance_uuids,
                                 counter_names):
        """Gets more cDOT performance counters."""

        # NOTE(nahimsouza): This conversion is nedeed because different names
        # are used in ZAPI and we want to avoid changes in the driver for now.
        rest_counter_names = {
            'domain_busy': 'domain_busy_percent',
            'processor_elapsed_time': 'elapsed_time',
            'avg_processor_busy': 'average_processor_busy_percent',
        }

        zapi_counter_names = {
            'domain_busy_percent': 'domain_busy',
            'elapsed_time': 'processor_elapsed_time',
            'average_processor_busy_percent': 'avg_processor_busy',
        }

        for i in range(len(counter_names)):
            if counter_names[i] in rest_counter_names:
                counter_names[i] = rest_counter_names[counter_names[i]]

        query = {
            'id': '|'.join(instance_uuids),
            'counters.name': '|'.join(counter_names),
            'fields': 'id,counter_table.name,counters.*',
        }

        response = self.send_request(
            f'/cluster/counter/tables/{object_name}/rows',
            'get', query=query, enable_tunneling=False)

        counter_data = []
        for record in response.get('records', []):
            for counter in record['counters']:

                counter_name = counter['name']

                # Reverts the name conversion
                if counter_name in zapi_counter_names:
                    counter_name = zapi_counter_names[counter_name]

                counter_value = ''
                if counter.get('value'):
                    counter_value = counter.get('value')
                elif counter.get('values'):
                    # NOTE(nahimsouza): Conversion made to keep compatibility
                    # with old ZAPI format
                    values = counter.get('values')
                    counter_value = ','.join([str(v) for v in values])

                counter_data.append({
                    'instance-name': record['counter_table']['name'],
                    'instance-uuid': record['id'],
                    'node-name': record['id'].split(':')[0],
                    'timestamp': int(time()),
                    counter_name: counter_value,
                })

        return counter_data

    def get_aggregate_capacities(self, aggregate_names):
        """Gets capacity info for multiple aggregates."""

        if not isinstance(aggregate_names, list):
            return {}

        aggregates = {}
        for aggregate_name in aggregate_names:
            aggregates[aggregate_name] = self._get_aggregate_capacity(
                aggregate_name)

        return aggregates

    def _get_aggregate_capacity(self, aggregate_name):
        """Gets capacity info for an aggregate."""

        fields = ('space.block_storage.available,space.block_storage.size,'
                  'space.block_storage.used')

        try:
            aggrs = self._get_aggregates(aggregate_names=[aggregate_name],
                                         fields=fields)

            result = {}
            if len(aggrs) > 0:
                aggr = aggrs[0]

                available = float(aggr['space']['block_storage']['available'])
                total = float(aggr['space']['block_storage']['size'])
                used = float(aggr['space']['block_storage']['used'])
                percent_used = int((used * 100) // total)

                result = {
                    'percent-used': percent_used,
                    'size-available': available,
                    'size-total': total,
                }

            return result
        except netapp_api.NaApiError as e:
            if (e.code == netapp_api.REST_API_NOT_FOUND or
                    e.code == netapp_api.REST_UNAUTHORIZED):
                LOG.debug('Aggregate capacity can only be collected with '
                          'cluster scoped credentials.')
            else:
                LOG.exception('Failed to get info for aggregate %s.',
                              aggregate_name)
            return {}

    def get_node_for_aggregate(self, aggregate_name):
        """Get home node for the specified aggregate.

        This API could return None, most notably if it was sent
        to a Vserver LIF, so the caller must be able to handle that case.
        """

        if not aggregate_name:
            return None

        fields = 'home_node.name'
        try:
            aggrs = self._get_aggregates(aggregate_names=[aggregate_name],
                                         fields=fields)
            node = None
            if len(aggrs) > 0:
                aggr = aggrs[0]
                node = aggr['home_node']['name']

            return node
        except netapp_api.NaApiError as e:
            if e.code == netapp_api.REST_API_NOT_FOUND:
                return None
            else:
                raise e

    def provision_qos_policy_group(self, qos_policy_group_info,
                                   qos_min_support):
        """Create QoS policy group on the backend if appropriate."""
        if qos_policy_group_info is None:
            return

        # Legacy QoS uses externally provisioned QoS policy group,
        # so we don't need to create one on the backend.
        legacy = qos_policy_group_info.get('legacy')
        if legacy:
            return

        spec = qos_policy_group_info.get('spec')

        if not spec:
            return

        is_adaptive = na_utils.is_qos_policy_group_spec_adaptive(
            qos_policy_group_info)
        self._validate_qos_policy_group(is_adaptive, spec=spec,
                                        qos_min_support=qos_min_support)

        qos_policy_group = self._get_qos_first_policy_group_by_name(
            spec['policy_name'])

        if not qos_policy_group:
            self._create_qos_policy_group(spec, is_adaptive)
        else:
            self._modify_qos_policy_group(spec, is_adaptive,
                                          qos_policy_group)

    def _get_qos_first_policy_group_by_name(self, qos_policy_group_name):
        records = self._get_qos_policy_group_by_name(qos_policy_group_name)
        if len(records) == 0:
            return None

        return records[0]

    def _get_qos_policy_group_by_name(self, qos_policy_group_name):
        query = {'name': qos_policy_group_name}

        response = self.send_request('/storage/qos/policies/',
                                     'get', query=query)

        records = response.get('records')
        if not records:
            return []

        return records

    def _qos_spec_to_api_args(self, spec, is_adaptive, vserver=None):
        """Convert a QoS spec to REST args."""
        rest_args = {}
        if is_adaptive:
            rest_args['adaptive'] = {}
            if spec.get('absolute_min_iops'):
                rest_args['adaptive']['absolute_min_iops'] = (
                    self._sanitize_qos_spec_value(
                        spec.get('absolute_min_iops')))
            if spec.get('expected_iops'):
                rest_args['adaptive']['expected_iops'] = (
                    self._sanitize_qos_spec_value(spec.get('expected_iops')))
            if spec.get('expected_iops_allocation'):
                rest_args['adaptive']['expected_iops_allocation'] = (
                    spec.get('expected_iops_allocation'))
            if spec.get('peak_iops'):
                rest_args['adaptive']['peak_iops'] = (
                    self._sanitize_qos_spec_value(spec.get('peak_iops')))
            if spec.get('peak_iops_allocation'):
                rest_args['adaptive']['peak_iops_allocation'] = (
                    spec.get('peak_iops_allocation'))
            if spec.get('block_size'):
                rest_args['adaptive']['block_size'] = (
                    spec.get('block_size'))
        else:
            rest_args['fixed'] = {}
            qos_max = spec.get('max_throughput')
            if qos_max and 'iops' in qos_max:
                rest_args['fixed']['max_throughput_iops'] = (
                    self._sanitize_qos_spec_value(qos_max))
            elif qos_max:
                # Convert from B/s to MB/s
                value = math.ceil(
                    self._sanitize_qos_spec_value(qos_max) / (10**6))
                rest_args['fixed']['max_throughput_mbps'] = value

            qos_min = spec.get('min_throughput')
            if qos_min and 'iops' in qos_min:
                rest_args['fixed']['min_throughput_iops'] = (
                    self._sanitize_qos_spec_value(qos_min))

        if spec.get('policy_name'):
            rest_args['name'] = spec.get('policy_name')
        if spec.get('return_record'):
            rest_args['return_records'] = spec.get('return_record')

        if vserver:
            rest_args['svm'] = {}
            rest_args['svm']['name'] = vserver

        return rest_args

    def _sanitize_qos_spec_value(self, value):
        value = value.lower()
        value = value.replace('iops', '').replace('b/s', '')
        value = int(value)
        return value

    def _create_qos_policy_group(self, spec, is_adaptive):
        """Creates a QoS policy group."""
        body = self._qos_spec_to_api_args(
            spec, is_adaptive, vserver=self.vserver)

        self.send_request('/storage/qos/policies/', 'post', body=body,
                          enable_tunneling=False)

    def _modify_qos_policy_group(self, spec, is_adaptive, qos_policy_group):
        """Modifies a QoS policy group."""
        body = self._qos_spec_to_api_args(spec, is_adaptive)
        if qos_policy_group['name'] == body['name']:
            body.pop('name')

        self.send_request(
            f'/storage/qos/policies/{qos_policy_group["uuid"]}', 'patch',
            body=body, enable_tunneling=False)

    def get_vol_by_junc_vserver(self, vserver, junction):
        """Gets the volume by junction path and vserver."""
        volume = self._get_volume_by_args(vol_path=junction, vserver=vserver)
        return volume['name']

    def file_assign_qos(self, flex_vol, qos_policy_group_name,
                        qos_policy_group_is_adaptive, file_path):
        """Assigns the named QoS policy-group to a file."""
        volume = self._get_volume_by_args(flex_vol)
        body = {
            'qos_policy.name': qos_policy_group_name
        }

        self.send_request(
            f'/storage/volumes/{volume["uuid"]}/files/{file_path}',
            'patch', body=body, enable_tunneling=False)

    def mark_qos_policy_group_for_deletion(self, qos_policy_group_info,
                                           is_adaptive=False):
        """Soft delete a QoS policy group backing a cinder volume."""
        if qos_policy_group_info is None:
            return

        spec = qos_policy_group_info.get('spec')

        # For cDOT we want to delete the QoS policy group that we created for
        # this cinder volume.  Because the QoS policy may still be "in use"
        # after the zapi call to delete the volume itself returns successfully,
        # we instead rename the QoS policy group using a specific pattern and
        # later attempt on a best effort basis to delete any QoS policy groups
        # matching that pattern.
        if spec:
            current_name = spec['policy_name']
            new_name = DELETED_PREFIX + current_name
            try:
                self._rename_qos_policy_group(current_name, new_name)
            except netapp_api.NaApiError as ex:
                LOG.warning('Rename failure in cleanup of cDOT QoS policy '
                            'group %(current_name)s: %(ex)s',
                            {'current_name': current_name, 'ex': ex})

        # Attempt to delete any QoS policies named "delete-openstack-*".
        self.remove_unused_qos_policy_groups()

    def delete_file(self, path_to_file):
        """Delete file at path."""
        LOG.debug('Deleting file: %s', path_to_file)

        volume_name = path_to_file.split('/')[2]
        relative_path = '/'.join(path_to_file.split('/')[3:])
        volume = self._get_volume_by_args(volume_name)

        # Path requires "%2E" to represent "." and "%2F" to represent "/".
        relative_path = relative_path.replace('.', '%2E').replace('/', '%2F')

        self.send_request(f'/storage/volumes/{volume["uuid"]}'
                          + f'/files/{relative_path}', 'delete')

    def _rename_qos_policy_group(self, qos_policy_group_name, new_name):
        """Renames a QoS policy group."""
        body = {'name': new_name}
        query = {'name': qos_policy_group_name}
        self.send_request('/storage/qos/policies/', 'patch', body=body,
                          query=query, enable_tunneling=False)

    def remove_unused_qos_policy_groups(self):
        """Deletes all QoS policy groups that are marked for deletion."""
        query = {'name': f'{DELETED_PREFIX}*'}
        self.send_request('/storage/qos/policies', 'delete', query=query)

    def create_lun(self, volume_name, lun_name, size, metadata,
                   qos_policy_group_name=None,
                   qos_policy_group_is_adaptive=False):
        """Issues API request for creating LUN on volume."""
        self._validate_qos_policy_group(qos_policy_group_is_adaptive)

        path = f'/vol/{volume_name}/{lun_name}'
        space_reservation = metadata['SpaceReserved']
        initial_size = size

        body = {
            'name': path,
            'space.size': str(initial_size),
            'os_type': metadata['OsType'],
            'space.guarantee.requested': space_reservation
        }

        if qos_policy_group_name:
            body['qos_policy.name'] = qos_policy_group_name

        try:
            self.send_request('/storage/luns', 'post', body=body)
        except netapp_api.NaApiError as ex:
            with excutils.save_and_reraise_exception():
                LOG.error('Error provisioning volume %(lun_name)s on '
                          '%(volume_name)s. Details: %(ex)s',
                          {
                              'lun_name': lun_name,
                              'volume_name': volume_name,
                              'ex': ex,
                          })

    def do_direct_resize(self, path, new_size_bytes, force=True):
        """Resize the LUN."""
        seg = path.split("/")
        LOG.info('Resizing LUN %s directly to new size.', seg[-1])

        body = {'name': path, 'space.size': new_size_bytes}

        self._lun_update_by_path(path, body)

    def _get_lun_by_path(self, path, fields=None):
        query = {'name': path}

        if fields:
            query['fields'] = fields

        response = self.send_request('/storage/luns', 'get', query=query)
        records = response.get('records', [])

        return records

    def _get_first_lun_by_path(self, path, fields=None):
        records = self._get_lun_by_path(path, fields=fields)
        if len(records) == 0:
            return None

        return records[0]

    def _lun_update_by_path(self, path, body):
        """Update the LUN."""
        lun = self._get_first_lun_by_path(path)

        if not lun:
            raise netapp_api.NaApiError(code=netapp_api.EOBJECTNOTFOUND)

        self.send_request(f'/storage/luns/{lun["uuid"]}', 'patch', body=body)

    def _validate_qos_policy_group(self, is_adaptive, spec=None,
                                   qos_min_support=False):
        if is_adaptive and not self.features.ADAPTIVE_QOS:
            msg = _("Adaptive QoS feature requires ONTAP 9.4 or later.")
            raise na_utils.NetAppDriverException(msg)

        if not spec:
            return

        if 'min_throughput' in spec and not qos_min_support:
            msg = 'min_throughput is not supported by this back end.'
            raise na_utils.NetAppDriverException(msg)

    def get_if_info_by_ip(self, ip):
        """Gets the network interface info by ip."""
        query_args = {}
        query_args['ip.address'] = volume_utils.resolve_hostname(ip)
        query_args['fields'] = 'svm'

        result = self.send_request('/network/ip/interfaces/', 'get',
                                   query=query_args, enable_tunneling=False)
        num_records = result['num_records']
        records = result.get('records', [])

        if num_records == 0:
            raise exception.NotFound(
                _('No interface found on cluster for ip %s') % ip)

        return [{'vserver': item['svm']['name']} for item in records]

    def get_igroup_by_initiators(self, initiator_list):
        """Get igroups exactly matching a set of initiators."""

        igroup_list = []
        if not initiator_list:
            return igroup_list

        query = {
            'svm.name': self.vserver,
            'initiators.name': ' '.join(initiator_list),
            'fields': 'name,protocol,os_type'
        }

        response = self.send_request('/protocols/san/igroups',
                                     'get', query=query)
        records = response.get('records', [])
        for igroup_item in records:
            igroup = {'initiator-group-os-type': igroup_item['os_type'],
                      'initiator-group-type': igroup_item['protocol'],
                      'initiator-group-name': igroup_item['name']}
            igroup_list.append(igroup)

        return igroup_list

    def add_igroup_initiator(self, igroup, initiator):
        """Adds initiators to the specified igroup."""
        query_initiator_uuid = {
            'name': igroup,
            'fields': 'uuid'
        }

        response_initiator_uuid = self.send_request(
            '/protocols/san/igroups/', 'get', query=query_initiator_uuid)

        response = response_initiator_uuid.get('records', [])
        if len(response) < 1:
            msg = _('Could not find igroup initiator.')
            raise exception.VolumeBackendAPIException(data=msg)

        igroup_uuid = response[0]['uuid']

        body = {
            'name': initiator
        }

        self.send_request('/protocols/san/igroups/' +
                          igroup_uuid + '/initiators',
                          'post', body=body)

    def create_igroup(self, igroup, igroup_type='iscsi', os_type='default'):
        """Creates igroup with specified args."""
        body = {
            'name': igroup,
            'protocol': igroup_type,
            'os_type': os_type,
        }
        self.send_request('/protocols/san/igroups', 'post', body=body)

    def map_lun(self, path, igroup_name, lun_id=None):
        """Maps LUN to the initiator and returns LUN id assigned."""

        body_post = {
            'lun.name': path,
            'igroup.name': igroup_name,
        }

        if lun_id is not None:
            body_post['logical_unit_number'] = lun_id

        try:
            result = self.send_request('/protocols/san/lun-maps', 'post',
                                       body=body_post,
                                       query={'return_records': 'true'})
            records = result.get('records')
            lun_id_assigned = records[0].get('logical_unit_number')
            return lun_id_assigned
        except netapp_api.NaApiError as e:
            code = e.code
            message = e.message
            LOG.warning('Error mapping LUN. Code :%(code)s, Message: '
                        '%(message)s', {'code': code, 'message': message})
            raise

    def get_lun_map(self, path):
        """Gets the LUN map by LUN path."""
        map_list = []

        query = {
            'lun.name': path,
            'fields': 'igroup.name,logical_unit_number,svm.name',
        }

        response = self.send_request('/protocols/san/lun-maps',
                                     'get',
                                     query=query)
        num_records = response.get('num_records')
        records = response.get('records', None)
        if records is None or num_records is None:
            return map_list

        for element in records:
            map_lun = {}
            map_lun['initiator-group'] = element['igroup']['name']
            map_lun['lun-id'] = element['logical_unit_number']
            map_lun['vserver'] = element['svm']['name']
            map_list.append(map_lun)

        return map_list

    def get_fc_target_wwpns(self):
        """Gets the FC target details."""
        wwpns = []
        query = {
            'fields': 'wwpn'
        }
        response = self.send_request('/network/fc/interfaces',
                                     'get', query=query)

        records = response.get('records')
        for record in records:
            wwpn = record.get('wwpn').lower()
            wwpns.append(wwpn)

        return wwpns

    def unmap_lun(self, path, igroup_name):
        """Unmaps a LUN from given initiator."""

        # get lun amd igroup uuids
        query_uuid = {
            'igroup.name': igroup_name,
            'lun.name': path,
            'fields': 'lun.uuid,igroup.uuid'
        }

        response_uuid = self.send_request(
            '/protocols/san/lun-maps', 'get', query=query_uuid)

        if response_uuid['num_records'] > 0:
            lun_uuid = response_uuid['records'][0]['lun']['uuid']
            igroup_uuid = response_uuid['records'][0]['igroup']['uuid']

            try:
                self.send_request(
                    f'/protocols/san/lun-maps/{lun_uuid}/{igroup_uuid}',
                    'delete')
            except netapp_api.NaApiError as e:
                LOG.warning("Error unmapping LUN. Code: %(code)s, Message: "
                            "%(message)s", {'code': e.code,
                                            'message': e.message})
                # if the LUN is already unmapped
                if e.code == netapp_api.REST_NO_SUCH_LUN_MAP:
                    pass
                else:
                    raise e
        else:
            # Input is invalid or LUN may already be unmapped
            LOG.warning("Error unmapping LUN. Invalid input.")

    def has_luns_mapped_to_initiators(self, initiator_list):
        """Checks whether any LUNs are mapped to the given initiator(s)."""
        query = {
            'initiators.name': ' '.join(initiator_list),
            'fields': 'lun_maps'
        }

        response = self.send_request('/protocols/san/igroups',
                                     'get', query=query)

        records = response.get('records', [])
        if len(records) > 0:
            for record in records:
                lun_maps = record.get('lun_maps', [])
                if len(lun_maps) > 0:
                    return True

        return False

    def get_iscsi_service_details(self):
        """Returns iscsi iqn."""
        query = {
            'fields': 'target.name'
        }
        response = self.send_request(
            '/protocols/san/iscsi/services', 'get', query=query)
        records = response.get('records')
        if records:
            return records[0]['target']['name']

        LOG.debug('No iSCSI service found for vserver %s', self.vserver)
        return None

    def check_iscsi_initiator_exists(self, iqn):
        """Returns True if initiator exists."""
        endpoint_url = '/protocols/san/iscsi/credentials'
        initiator_exists = True
        try:
            query = {
                'initiator': iqn,
            }
            response = self.send_request(endpoint_url, 'get', query=query)
            records = response.get('records')
            if not records:
                initiator_exists = False

        except netapp_api.NaApiError:
            initiator_exists = False

        return initiator_exists

    def set_iscsi_chap_authentication(self, iqn, username, password):
        """Provides NetApp host's CHAP credentials to the backend."""
        initiator_exists = self.check_iscsi_initiator_exists(iqn)

        command_template = ('iscsi security %(mode)s -vserver %(vserver)s '
                            '-initiator-name %(iqn)s -auth-type CHAP '
                            '-user-name %(username)s')

        if initiator_exists:
            LOG.debug('Updating CHAP authentication for %(iqn)s.',
                      {'iqn': iqn})
            command = command_template % {
                'mode': 'modify',
                'vserver': self.vserver,
                'iqn': iqn,
                'username': username,
            }
        else:
            LOG.debug('Adding initiator %(iqn)s with CHAP authentication.',
                      {'iqn': iqn})
            command = command_template % {
                'mode': 'create',
                'vserver': self.vserver,
                'iqn': iqn,
                'username': username,
            }

        try:
            with self.ssh_client.ssh_connect_semaphore:
                ssh_pool = self.ssh_client.ssh_pool
                with ssh_pool.item() as ssh:
                    self.ssh_client.execute_command_with_prompt(ssh,
                                                                command,
                                                                'Password:',
                                                                password)
        except Exception as e:
            msg = _('Failed to set CHAP authentication for target IQN %(iqn)s.'
                    ' Details: %(ex)s') % {
                'iqn': iqn,
                'ex': e,
            }
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def get_iscsi_target_details(self):
        """Gets the iSCSI target portal details."""
        query = {
            'services': 'data_iscsi',
            'fields': 'ip.address,enabled'
        }

        response = self.send_request('/network/ip/interfaces',
                                     'get', query=query)

        target_list = []
        records = response.get('records', [])
        for record in records:
            details = dict()
            details['address'] = record['ip']['address']
            details['tpgroup-tag'] = None
            details['interface-enabled'] = record['enabled']
            # NOTE(nahimsouza): from ONTAP documentation:
            # ONTAP does not support changing the port number for iSCSI.
            # Port number 3260 is registered as part of the iSCSI specification
            # and cannot be used by any other application or service.
            details['port'] = 3260
            target_list.append(details)

        return target_list

    def move_lun(self, path, new_path):
        """Moves the LUN at path to new path."""
        seg = path.split("/")
        new_seg = new_path.split("/")
        LOG.debug("Moving LUN %(name)s to %(new_name)s.",
                  {'name': seg[-1], 'new_name': new_seg[-1]})
        query = {
            'svm.name': self.vserver,
            'name': path
        }
        body = {
            'name': new_path,
        }
        self.send_request('/storage/luns/', 'patch', query=query, body=body)

    def clone_file(self, flex_vol, src_path, dest_path, vserver,
                   dest_exists=False, source_snapshot=None, is_snapshot=False):
        """Clones file on vserver."""
        LOG.debug('Cloning file - volume %(flex_vol)s, src %(src_path)s, '
                  'dest %(dest_path)s, vserver %(vserver)s,'
                  'source_snapshot %(source_snapshot)s',
                  {
                      'flex_vol': flex_vol,
                      'src_path': src_path,
                      'dest_path': dest_path,
                      'vserver': vserver,
                      'source_snapshot': source_snapshot,
                  })

        volume = self._get_volume_by_args(flex_vol)
        body = {
            'volume': {
                'uuid': volume['uuid'],
                'name': volume['name']
            },
            'source_path': src_path,
            'destination_path': dest_path,
        }
        if is_snapshot and self.features.BACKUP_CLONE_PARAM:
            body['is_backup'] = True

        if dest_exists:
            body['overwrite_destination'] = True

        self.send_request('/storage/file/clone', 'post', body=body)

    def clone_lun(self, volume, name, new_name, space_reserved='true',
                  qos_policy_group_name=None, src_block=0, dest_block=0,
                  block_count=0, source_snapshot=None, is_snapshot=False,
                  qos_policy_group_is_adaptive=False):
        """Clones lun on vserver."""
        LOG.debug('Cloning lun - volume: %(volume)s, name: %(name)s, '
                  'new_name: %(new_name)s, space_reserved: %(space_reserved)s,'
                  ' qos_policy_group_name: %(qos_policy_group_name)s',
                  {
                      'volume': volume,
                      'name': name,
                      'new_name': new_name,
                      'space_reserved': space_reserved,
                      'qos_policy_group_name': qos_policy_group_name,
                  })

        # NOTE(nahimsouza): some parameters are not available on REST API,
        # but they are in the header just to keep compatilbility with ZAPI:
        # src_block, dest_block, block_count, is_snapshot

        self._validate_qos_policy_group(qos_policy_group_is_adaptive)

        source_path = f'/vol/{volume}'
        if source_snapshot:
            source_path += f'/.snapshot/{source_snapshot}'
        source_path += f'/{name}'
        body = {
            'svm': {
                'name': self.vserver
            },
            'name': f'/vol/{volume}/{new_name}',
            'clone': {
                'source': {
                    'name': source_path,
                }
            },
            'space': {
                'guarantee': {
                    'requested': space_reserved == 'true',
                }
            }
        }

        if qos_policy_group_name:
            body['qos_policy'] = {'name': qos_policy_group_name}

        self.send_request('/storage/luns', 'post', body=body)

    def destroy_lun(self, path, force=True):
        """Destroys the LUN at the path."""
        query = {}
        query['name'] = path
        query['svm'] = self.vserver

        if force:
            query['allow_delete_while_mapped'] = 'true'

        self.send_request('/storage/luns/', 'delete', query=query)

    def get_flexvol_capacity(self, flexvol_path=None, flexvol_name=None):
        """Gets total capacity and free capacity, in bytes, of the flexvol."""
        fields = 'name,space.available,space.afs_total'
        try:
            volume = self._get_volume_by_args(
                vol_name=flexvol_name, vol_path=flexvol_path, fields=fields)
            capacity = {
                'size-total': float(volume['space']['afs_total']),
                'size-available': float(volume['space']['available']),
            }
            return capacity
        except exception.VolumeBackendAPIException:
            msg = _('Volume %s not found.')
            msg_args = flexvol_path or flexvol_name
            raise na_utils.NetAppDriverException(msg % msg_args)

    def get_provisioning_options_from_flexvol(self, flexvol_name):
        """Get a dict of provisioning options matching existing flexvol."""

        flexvol_info = self.get_flexvol(flexvol_name=flexvol_name)
        dedupe_info = self.get_flexvol_dedupe_info(flexvol_name)

        provisioning_opts = {
            'aggregate': flexvol_info['aggregate'],
            # space-guarantee can be 'none', 'file', 'volume'
            'space_guarantee_type': flexvol_info.get('space-guarantee'),
            'snapshot_policy': flexvol_info['snapshot-policy'],
            'language': flexvol_info['language'],
            'dedupe_enabled': dedupe_info['dedupe'],
            'compression_enabled': dedupe_info['compression'],
            'snapshot_reserve': flexvol_info['percentage-snapshot-reserve'],
            'volume_type': flexvol_info['type'],
            'size': int(math.ceil(float(flexvol_info['size']) / units.Gi)),
            'is_flexgroup': flexvol_info['style-extended'] == 'flexgroup',
        }

        return provisioning_opts

    def flexvol_exists(self, volume_name):
        """Checks if a flexvol exists on the storage array."""
        LOG.debug('Checking if volume %s exists', volume_name)

        query = {
            'name': volume_name,
            'return_records': 'false'
        }

        response = self.send_request('/storage/volumes/', 'get', query=query)

        return response['num_records'] > 0

    def create_volume_async(self, name, aggregate_list, size_gb,
                            space_guarantee_type=None, snapshot_policy=None,
                            language=None, dedupe_enabled=False,
                            compression_enabled=False, snapshot_reserve=None,
                            volume_type='rw'):
        """Creates a volume asynchronously."""

        body = {
            'name': name,
            'size': size_gb * units.Gi,
            'type': volume_type,
        }

        if isinstance(aggregate_list, list):
            body['style'] = 'flexgroup'
            body['aggregates'] = [{'name': aggr} for aggr in aggregate_list]
        else:
            body['style'] = 'flexvol'
            body['aggregates'] = [{'name': aggregate_list}]

        if volume_type == 'dp':
            snapshot_policy = None
        else:
            body['nas'] = {'path': '/%s' % name}

        if snapshot_policy is not None:
            body['snapshot_policy'] = {'name': snapshot_policy}

        if space_guarantee_type:
            body['guarantee'] = {'type': space_guarantee_type}

        if language is not None:
            body['language'] = language

        if snapshot_reserve is not None:
            body['space'] = {
                'snapshot': {
                    'reserve_percent': str(snapshot_reserve)
                }
            }

        # cDOT compression requires that deduplication be enabled.
        if dedupe_enabled or compression_enabled:
            body['efficiency'] = {'dedupe': 'background'}

        if compression_enabled:
            body['efficiency']['compression'] = 'background'

        response = self.send_request('/storage/volumes/', 'post', body=body,
                                     wait_on_accepted=False)

        job_info = {
            'status': None,
            'jobid': response["job"]["uuid"],
            'error-code': None,
            'error-message': None,
        }

        return job_info

    def create_flexvol(self, flexvol_name, aggregate_name, size_gb,
                       space_guarantee_type=None, snapshot_policy=None,
                       language=None, dedupe_enabled=False,
                       compression_enabled=False, snapshot_reserve=None,
                       volume_type='rw'):
        """Creates a flexvol asynchronously and return the job info."""

        return self.create_volume_async(
            flexvol_name, aggregate_name, size_gb,
            space_guarantee_type=space_guarantee_type,
            snapshot_policy=snapshot_policy, language=language,
            dedupe_enabled=dedupe_enabled,
            compression_enabled=compression_enabled,
            snapshot_reserve=snapshot_reserve, volume_type=volume_type)

    def enable_volume_dedupe_async(self, volume_name):
        """Enable deduplication on FlexVol/FlexGroup volume asynchronously."""

        query = {
            'name': volume_name,
            'fields': 'uuid,style',
        }
        body = {
            'efficiency': {'dedupe': 'background'}
        }
        self.send_request('/storage/volumes/', 'patch', body=body, query=query,
                          wait_on_accepted=False)

    def enable_volume_compression_async(self, volume_name):
        """Enable compression on FlexVol/FlexGroup volume asynchronously."""
        query = {
            'name': volume_name
        }
        body = {
            'efficiency': {'compression': 'background'}
        }
        self.send_request('/storage/volumes/', 'patch', body=body, query=query,
                          wait_on_accepted=False)

    def _parse_lagtime(self, time_str):
        """Parse lagtime string (ISO 8601) into a number of seconds."""

        fmt_str = 'PT'
        if 'H' in time_str:
            fmt_str += '%HH'
        if 'M' in time_str:
            fmt_str += '%MM'
        if 'S' in time_str:
            fmt_str += '%SS'

        t = None
        try:
            t = datetime.strptime(time_str, fmt_str)
        except Exception:
            LOG.debug("Failed to parse lagtime: %s", time_str)
            raise

        # convert to timedelta to get the total seconds
        td = timedelta(hours=t.hour, minutes=t.minute, seconds=t.second)
        return td.total_seconds()

    def _get_snapmirrors(self, source_vserver=None, source_volume=None,
                         destination_vserver=None, destination_volume=None):

        fields = ['state', 'source.svm.name', 'source.path',
                  'destination.svm.name', 'destination.path',
                  'transfer.end_time', 'lag_time', 'healthy', 'uuid']

        query = {}
        query['fields'] = '{}'.format(','.join(f for f in fields))

        query_src_vol = source_volume if source_volume else '*'
        query_src_vserver = source_vserver if source_vserver else '*'
        query['source.path'] = query_src_vserver + ':' + query_src_vol

        query_dst_vol = destination_volume if destination_volume else '*'
        query_dst_vserver = destination_vserver if destination_vserver else '*'
        query['destination.path'] = query_dst_vserver + ':' + query_dst_vol

        response = self.send_request(
            '/snapmirror/relationships', 'get', query=query)

        snapmirrors = []
        for record in response.get('records', []):
            snapmirrors.append({
                'relationship-status': record.get('state'),
                'mirror-state': record['state'],
                'source-vserver': record['source']['svm']['name'],
                'source-volume': (record['source']['path'].split(':')[1] if
                                  record.get('source') else None),
                'destination-vserver': record['destination']['svm']['name'],
                'destination-volume': (
                    record['destination']['path'].split(':')[1]
                    if record.get('destination') else None),
                'last-transfer-end-timestamp':
                    (record['transfer']['end_time'] if
                     record.get('transfer', {}).get('end_time') else None),
                'lag-time': (self._parse_lagtime(record['lag_time']) if
                             record.get('lag_time') else None),
                'is-healthy': record['healthy'],
                'uuid': record['uuid']
            })

        return snapmirrors

    def get_snapmirrors(self, source_vserver, source_volume,
                        destination_vserver, destination_volume,
                        desired_attributes=None):
        """Gets one or more SnapMirror relationships.

        Either the source or destination info may be omitted.
        Desired attributes exists only to keep consistent with ZAPI client
        signature and has no effect in the output.
        """

        snapmirrors = self._get_snapmirrors(
            source_vserver=source_vserver,
            source_volume=source_volume,
            destination_vserver=destination_vserver,
            destination_volume=destination_volume)

        return snapmirrors

    def create_snapmirror(self, source_vserver, source_volume,
                          destination_vserver, destination_volume,
                          schedule=None, policy=None,
                          relationship_type='data_protection'):
        """Creates a SnapMirror relationship.

        The schedule and relationship type is kept to avoid breaking
        the API used by data_motion, but are not used on the REST API.

        The schedule is part of the policy associated the relationship and the
        relationship_type will be ignored because XDP is the only type
        supported through REST API.
        """

        body = {
            'source': {
                'path': source_vserver + ':' + source_volume
            },
            'destination': {
                'path': destination_vserver + ':' + destination_volume
            }
        }

        if policy:
            body['policy'] = {'name': policy}

        try:
            self.send_request('/snapmirror/relationships/', 'post', body=body)
        except netapp_api.NaApiError as e:
            if e.code != netapp_api.REST_ERELATION_EXISTS:
                raise e

    def _set_snapmirror_state(self, state, source_vserver, source_volume,
                              destination_vserver, destination_volume,
                              wait_result=True):
        """Change the snapmirror state between two volumes."""

        snapmirror = self.get_snapmirrors(source_vserver, source_volume,
                                          destination_vserver,
                                          destination_volume)

        if not snapmirror:
            msg = _('Failed to get information about relationship between '
                    'source %(src_vserver)s:%(src_volume)s and '
                    'destination %(dst_vserver)s:%(dst_volume)s.') % {
                'src_vserver': source_vserver,
                'src_volume': source_volume,
                'dst_vserver': destination_vserver,
                'dst_volume': destination_volume}
            raise na_utils.NetAppDriverException(msg)

        uuid = snapmirror[0]['uuid']
        body = {'state': state}
        result = self.send_request('/snapmirror/relationships/' + uuid,
                                   'patch', body=body,
                                   wait_on_accepted=wait_result)
        job = result['job']
        job_info = {
            'operation-id': None,
            'status': None,
            'jobid': job.get('uuid'),
            'error-code': None,
            'error-message': None,
            'relationship-uuid': uuid,
        }

        return job_info

    def initialize_snapmirror(self, source_vserver, source_volume,
                              destination_vserver, destination_volume,
                              source_snapshot=None, transfer_priority=None):
        """Initializes a SnapMirror relationship."""

        # TODO: Trigger a geometry exception to be caught by data_motion.
        # This error is raised when using ZAPI with different volume component
        # numbers, but in REST, the job must be checked sometimes before that
        # error occurs.

        return self._set_snapmirror_state(
            'snapmirrored', source_vserver, source_volume,
            destination_vserver, destination_volume, wait_result=False)

    def abort_snapmirror(self, source_vserver, source_volume,
                         destination_vserver, destination_volume,
                         clear_checkpoint=False):
        """Stops ongoing transfers for a SnapMirror relationship."""

        snapmirror = self.get_snapmirrors(source_vserver, source_volume,
                                          destination_vserver,
                                          destination_volume)
        if not snapmirror:
            msg = _('Failed to get information about relationship between '
                    'source %(src_vserver)s:%(src_volume)s and '
                    'destination %(dst_vserver)s:%(dst_volume)s.') % {
                'src_vserver': source_vserver,
                'src_volume': source_volume,
                'dst_vserver': destination_vserver,
                'dst_volume': destination_volume}
            raise na_utils.NetAppDriverException(msg)

        snapmirror_uuid = snapmirror[0]['uuid']

        query = {'state': 'transferring'}
        transfers = self.send_request('/snapmirror/relationships/' +
                                      snapmirror_uuid + '/transfers/', 'get',
                                      query=query)

        if not transfers.get('records'):
            raise netapp_api.NaApiError(
                code=netapp_api.ENOTRANSFER_IN_PROGRESS)

        body = {'state': 'hard_aborted' if clear_checkpoint else 'aborted'}

        for transfer in transfers['records']:
            transfer_uuid = transfer['uuid']
            self.send_request('/snapmirror/relationships/' +
                              snapmirror_uuid + '/transfers/' +
                              transfer_uuid, 'patch', body=body)

    def delete_snapmirror(self, source_vserver, source_volume,
                          destination_vserver, destination_volume):

        """Deletes an SnapMirror relationship on destination."""

        query_uuid = {}
        query_uuid['source.path'] = source_vserver + ':' + source_volume
        query_uuid['destination.path'] = (destination_vserver + ':' +
                                          destination_volume)
        query_uuid['fields'] = 'uuid'

        response = self.send_request('/snapmirror/relationships/', 'get',
                                     query=query_uuid)

        records = response.get('records')
        if not records:
            raise netapp_api.NaApiError(code=netapp_api.EOBJECTNOTFOUND)

        # 'destination_only' deletes the snapmirror on destination but does not
        # release it on source.
        query_delete = {"destination_only": "true"}

        snapmirror_uuid = records[0].get('uuid')
        self.send_request('/snapmirror/relationships/' +
                          snapmirror_uuid, 'delete',
                          query=query_delete)

    def resume_snapmirror(self, source_vserver, source_volume,
                          destination_vserver, destination_volume):

        """Resume a SnapMirror relationship."""

        query_uuid = {}
        query_uuid['source.path'] = source_vserver + ':' + source_volume
        query_uuid['destination.path'] = (destination_vserver + ':' +
                                          destination_volume)
        query_uuid['fields'] = 'uuid,policy.type'

        response_snapmirrors = self.send_request('/snapmirror/relationships/',
                                                 'get', query=query_uuid)

        records = response_snapmirrors.get('records')
        if not records:
            raise netapp_api.NaApiError(code=netapp_api.EOBJECTNOTFOUND)

        snapmirror_uuid = records[0]['uuid']
        snapmirror_policy = records[0]['policy']['type']

        body_resync = {}
        if snapmirror_policy == 'async':
            body_resync['state'] = 'snapmirrored'
        elif snapmirror_policy == 'sync':
            body_resync['state'] = 'in_sync'

        self.send_request('/snapmirror/relationships/' +
                          snapmirror_uuid, 'patch',
                          body=body_resync)

    def release_snapmirror(self, source_vserver, source_volume,
                           destination_vserver, destination_volume,
                           relationship_info_only=False):
        """Removes a SnapMirror relationship on the source endpoint."""

        query_uuid = {}
        query_uuid['list_destinations_only'] = 'true'
        query_uuid['source.path'] = source_vserver + ':' + source_volume
        query_uuid['destination.path'] = (destination_vserver + ':' +
                                          destination_volume)
        query_uuid['fields'] = 'uuid'

        response_snapmirrors = self.send_request('/snapmirror/relationships/',
                                                 'get', query=query_uuid)

        records = response_snapmirrors.get('records')
        if not records:
            raise netapp_api.NaApiError(code=netapp_api.EOBJECTNOTFOUND)

        query_release = {}
        if relationship_info_only:
            # release without removing related snapshots
            query_release['source_info_only'] = 'true'
        else:
            # release and removing all related snapshots
            query_release['source_only'] = 'true'

        snapmirror_uuid = records[0].get('uuid')
        self.send_request('/snapmirror/relationships/' +
                          snapmirror_uuid, 'delete',
                          query=query_release)

    def resync_snapmirror(self, source_vserver, source_volume,
                          destination_vserver, destination_volume):
        """Resync a SnapMirror relationship."""

        # We reuse the resume operation for resync since both are handled in
        # the same way in the REST API, by setting the snapmirror relationship
        # to the snapmirrored state.
        self.resume_snapmirror(source_vserver,
                               source_volume,
                               destination_vserver,
                               destination_volume)

    def quiesce_snapmirror(self, source_vserver, source_volume,
                           destination_vserver, destination_volume):
        """Disables future transfers to a SnapMirror destination."""

        return self._set_snapmirror_state(
            'paused', source_vserver, source_volume,
            destination_vserver, destination_volume)

    def break_snapmirror(self, source_vserver, source_volume,
                         destination_vserver, destination_volume):
        """Breaks a data protection SnapMirror relationship."""

        self._set_snapmirror_state(
            'broken-off', source_vserver, source_volume,
            destination_vserver, destination_volume)

    def update_snapmirror(self, source_vserver, source_volume,
                          destination_vserver, destination_volume):
        """Schedules a SnapMirror update."""

        snapmirror = self.get_snapmirrors(source_vserver, source_volume,
                                          destination_vserver,
                                          destination_volume)
        if not snapmirror:
            msg = _('Failed to get information about relationship between '
                    'source %(src_vserver)s:%(src_volume)s and '
                    'destination %(dst_vserver)s:%(dst_volume)s.') % {
                'src_vserver': source_vserver,
                'src_volume': source_volume,
                'dst_vserver': destination_vserver,
                'dst_volume': destination_volume}

            raise na_utils.NetAppDriverException(msg)

        snapmirror_uuid = snapmirror[0]['uuid']

        # NOTE(nahimsouza): A POST with an empty body starts the update
        # snapmirror operation.
        try:
            self.send_request('/snapmirror/relationships/' +
                              snapmirror_uuid + '/transfers/', 'post',
                              wait_on_accepted=False)
        except netapp_api.NaApiError as e:
            if (e.code != netapp_api.REST_UPDATE_SNAPMIRROR_FAILED):
                LOG.warning('Unexpected failure during snapmirror update.'
                            'Code: %(code)s, Message: %(message)s',
                            {'code': e.code, 'message': e.message})
            raise

    def mount_flexvol(self, flexvol_name, junction_path=None):
        """Mounts a volume on a junction path."""
        query = {'name': flexvol_name}
        body = {'nas.path': (
            junction_path if junction_path else '/%s' % flexvol_name)}
        self.send_request('/storage/volumes', 'patch', query=query, body=body)

    def get_cluster_name(self):
        """Gets cluster name."""
        query = {'fields': 'name'}

        response = self.send_request('/cluster', 'get', query=query,
                                     enable_tunneling=False)

        return response['name']

    def get_vserver_peers(self, vserver_name=None, peer_vserver_name=None):
        """Gets one or more Vserver peer relationships."""
        query = {
            'fields': 'svm.name,state,peer.svm.name,peer.cluster.name,'
                      'applications'
        }

        if peer_vserver_name:
            query['name'] = peer_vserver_name
        if vserver_name:
            query['svm.name'] = vserver_name

        response = self.send_request('/svm/peers', 'get', query=query,
                                     enable_tunneling=False)
        records = response.get('records', [])

        vserver_peers = []
        for vserver_info in records:
            vserver_peer = {
                'vserver': vserver_info['svm']['name'],
                'peer-vserver': vserver_info['peer']['svm']['name'],
                'peer-state': vserver_info['state'],
                'peer-cluster': vserver_info['peer']['cluster']['name'],
                'applications': vserver_info['applications'],
            }
            vserver_peers.append(vserver_peer)

        return vserver_peers

    def create_vserver_peer(self, vserver_name, peer_vserver_name,
                            vserver_peer_application=None):
        """Creates a Vserver peer relationship."""
        # default peering application to `snapmirror` if none is specified.
        if not vserver_peer_application:
            vserver_peer_application = ['snapmirror']

        body = {
            'svm.name': vserver_name,
            'name': peer_vserver_name,
            'applications': vserver_peer_application
        }

        self.send_request('/svm/peers', 'post', body=body,
                          enable_tunneling=False)

    def start_lun_move(self, lun_name, dest_ontap_volume,
                       src_ontap_volume=None, dest_lun_name=None):
        """Starts a lun move operation between ONTAP volumes."""
        if dest_lun_name is None:
            dest_lun_name = lun_name
        if src_ontap_volume is None:
            src_ontap_volume = dest_ontap_volume

        src_path = f'/vol/{src_ontap_volume}/{lun_name}'
        dest_path = f'/vol/{dest_ontap_volume}/{dest_lun_name}'
        body = {'name': dest_path}
        self._lun_update_by_path(src_path, body)

        return dest_path

    def get_lun_move_status(self, dest_path):
        """Get lun move job status from a given dest_path."""
        lun = self._get_first_lun_by_path(
            dest_path, fields='movement.progress')

        if not lun:
            return None

        move_progress = lun['movement']['progress']
        move_status = {
            'job-status': move_progress['state'],
            'last-failure-reason': (move_progress
                                    .get('failure', {})
                                    .get('message', None))
        }

        return move_status

    def start_lun_copy(self, lun_name, dest_ontap_volume, dest_vserver,
                       src_ontap_volume=None, src_vserver=None,
                       dest_lun_name=None):
        """Starts a lun copy operation between ONTAP volumes."""
        if src_ontap_volume is None:
            src_ontap_volume = dest_ontap_volume
        if src_vserver is None:
            src_vserver = dest_vserver
        if dest_lun_name is None:
            dest_lun_name = lun_name

        src_path = f'/vol/{src_ontap_volume}/{lun_name}'
        dest_path = f'/vol/{dest_ontap_volume}/{dest_lun_name}'

        body = {
            'name': dest_path,
            'copy.source.name': src_path,
            'svm.name': dest_vserver
        }

        self.send_request('/storage/luns', 'post', body=body,
                          enable_tunneling=False)

        return dest_path

    def get_lun_copy_status(self, dest_path):
        """Get lun copy job status from a given dest_path."""
        lun = self._get_first_lun_by_path(
            dest_path, fields='copy.source.progress')

        if not lun:
            return None

        copy_progress = lun['copy']['source']['progress']
        copy_status = {
            'job-status': copy_progress['state'],
            'last-failure-reason': (copy_progress
                                    .get('failure', {})
                                    .get('message', None))
        }

        return copy_status

    def cancel_lun_copy(self, dest_path):
        """Cancel an in-progress lun copy by deleting the lun."""
        query = {
            'name': dest_path,
            'svm.name': self.vserver
        }

        try:
            self.send_request('/storage/luns/', 'delete', query=query)
        except netapp_api.NaApiError as e:
            msg = (_('Could not cancel lun copy by deleting lun at %s. %s'))
            raise na_utils.NetAppDriverException(msg % (dest_path, e))

    def start_file_copy(self, file_name, dest_ontap_volume,
                        src_ontap_volume=None,
                        dest_file_name=None):
        """Starts a file copy operation between ONTAP volumes."""
        if src_ontap_volume is None:
            src_ontap_volume = dest_ontap_volume
        if dest_file_name is None:
            dest_file_name = file_name

        source_vol = self._get_volume_by_args(src_ontap_volume)

        dest_vol = source_vol
        if dest_ontap_volume != src_ontap_volume:
            dest_vol = self._get_volume_by_args(dest_ontap_volume)

        body = {
            'files_to_copy': [
                {
                    'source': {
                        'path': f'{src_ontap_volume}/{file_name}',
                        'volume': {
                            'uuid': source_vol['uuid']
                        }
                    },
                    'destination': {
                        'path': f'{dest_ontap_volume}/{dest_file_name}',
                        'volume': {
                            'uuid': dest_vol['uuid']
                        }
                    }
                }
            ]
        }

        result = self.send_request('/storage/file/copy', 'post', body=body,
                                   enable_tunneling=False)
        return result['job']['uuid']

    def get_file_copy_status(self, job_uuid):
        """Get file copy job status from a given job's UUID."""
        # TODO(rfluisa): Select only the fields that are needed here.
        query = {}
        query['fields'] = '*'

        result = self.send_request(
            f'/cluster/jobs/{job_uuid}', 'get', query=query,
            enable_tunneling=False)

        if not result or not result.get('state', None):
            return None

        state = result.get('state')
        if state == 'success':
            state = 'complete'
        elif state == 'failure':
            state = 'destroyed'

        copy_status = {
            'job-status': state,
            'last-failure-reason': result.get('error', {}).get('message', None)
        }

        return copy_status

    def rename_file(self, orig_file_name, new_file_name):
        """Rename a volume file."""
        LOG.debug("Renaming the file %(original)s to %(new)s.",
                  {'original': orig_file_name, 'new': new_file_name})

        unique_volume = self._get_volume_by_args(
            vol_name=orig_file_name.split('/')[2])

        # Get the relative path
        orig_file_name = '/'.join(orig_file_name.split('/')[3:])
        new_file_name = '/'.join(new_file_name.split('/')[3:])

        # Path requires "%2E" to represent "." and "%2F" to represent "/".
        orig_file_name = orig_file_name.replace('.', '%2E').replace('/', '%2F')
        new_file_name = new_file_name.replace('.', '%2E').replace('/', '%2F')

        body = {'path': new_file_name}

        self.send_request(
            f'/storage/volumes/{unique_volume["uuid"]}/files/{orig_file_name}',
            'patch', body=body)



# Copyright (c) 2022 DataCore Software Corp. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Password storage."""

import contextlib
import json
import os
import stat

from oslo_log import log as logging

from cinder.i18n import _
from cinder import utils as cinder_utils

LOG = logging.getLogger(__name__)


class FileStorage(object):
    """Represents a file as a dictionary."""

    def __init__(self, file_path):
        self._file_path = file_path
        self._file = None
        self._is_open = False

    def open(self):
        """Open a file for simultaneous reading and writing.

        If the specified file does not exist, it will be created
        with the 0600 access permissions for the current user, if needed
        the appropriate directories will be created with the 0750 access
        permissions for the current user.
        """

        file_dir = os.path.dirname(self._file_path)
        if file_dir and not os.path.isdir(file_dir):
            os.makedirs(file_dir)
            os.chmod(file_dir, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP)
        if not os.path.isfile(self._file_path):
            open(self._file_path, 'w').close()
            os.chmod(self._file_path, stat.S_IRUSR | stat.S_IWUSR)

        if self._file:
            self.close()
        self._file = open(self._file_path, 'r+')
        return self

    def load(self):
        """Reads the file and returns corresponded dictionary object.

        :return: The dictionary that represents the file content.
        """

        storage = {}
        if os.stat(self._file_path).st_size != 0:
            storage = json.load(self._file)
            if not isinstance(storage, dict):
                msg = _('File %s has a malformed format.') % self._file_path
                raise ValueError(msg)
        return storage

    def save(self, storage):
        """Writes the specified dictionary to the file.

        :param storage: Dictionary that should be written to the file.
        """

        if not isinstance(storage, dict):
            msg = _('%s is not a dict.') % repr(storage)
            raise TypeError(msg)

        self._file.seek(0)
        self._file.truncate()
        json.dump(storage, self._file)

    def close(self):
        """Close the file."""

        if self._file:
            self._file.close()
        self._file = None


class PasswordFileStorage(object):
    """Password storage implementation.

    It stores passwords in a file in a clear text. The password file must be
    secured by setting up file permissions.
    """

    def __init__(self, file_path):
        self._file_path = file_path
        self._file_storage = FileStorage(file_path)

    def set_password(self, resource, username, password):
        """Store the credential for the resource.

        :param resource: Resource name for which credential will be stored
        :param username: User name
        :param password: Password
        """

        @cinder_utils.synchronized(
            'datacore-password_storage-' + self._file_path, external=True)
        def _set_password():
            with contextlib.closing(self._file_storage.open()) as storage:
                passwords = storage.load()
                if resource not in passwords:
                    passwords[resource] = {}
                passwords[resource][username] = password
                storage.save(passwords)

        _set_password()

    def get_password(self, resource, username):
        """Returns the stored password for the resource.

        If the password does not exist, it will return None

        :param resource: Resource name for which credential was stored
        :param username: User name
        :return password: Password
        """

        @cinder_utils.synchronized(
            'datacore-password_storage-' + self._file_path, external=True)
        def _get_password():
            with contextlib.closing(self._file_storage.open()) as storage:
                passwords = storage.load()
            if resource in passwords:
                return passwords[resource].get(username)

        return _get_password()

    def delete_password(self, resource, username):
        """Delete the stored credential for the resource.

        :param resource: Resource name for which credential was stored
        :param username: User name
        """

        @cinder_utils.synchronized(
            'datacore-password_storage-' + self._file_path, external=True)
        def _delete_password():
            with contextlib.closing(self._file_storage.open()) as storage:
                passwords = storage.load()
                if resource in passwords and username in passwords[resource]:
                    del passwords[resource][username]
                    if not passwords[resource].keys():
                        del passwords[resource]
                    storage.save(passwords)

        _delete_password()



# Copyright (c) 2016 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import unittest
from unittest import mock

import ddt
from oslo_utils import units

from cinder import coordination
from cinder.tests.unit.volume.drivers.dell_emc.unity \
    import fake_enum as enums
from cinder.tests.unit.volume.drivers.dell_emc.unity \
    import fake_exception as ex
from cinder.volume.drivers.dell_emc.unity import client


########################
#
#   Start of Mocks
#
########################


class MockResource(object):
    def __init__(self, name=None, _id=None):
        self.name = name
        self._id = _id
        self.existed = True
        self.size_total = 5 * units.Gi
        self.size_subscribed = 6 * units.Gi
        self.size_free = 2 * units.Gi
        self.is_auto_delete = None
        self.initiator_id = []
        self.alu_hlu_map = {'already_attached': 99}
        self.ip_address = None
        self.is_logged_in = None
        self.wwn = None
        self.max_iops = None
        self.max_kbps = None
        self.pool_name = 'Pool0'
        self._storage_resource = None
        self.host_cache = []
        self.is_thin = None
        self.is_all_flash = True
        self.description = None
        self.luns = None
        self.lun = None
        self.tiering_policy = None
        self.pool_fast_vp = None
        self.snap = True

    @property
    def id(self):
        return self._id

    def get_id(self):
        return self._id

    def delete(self, force_snap_delete=None):
        if self.get_id() in ['snap_2']:
            raise ex.SnapDeleteIsCalled()
        elif self.get_id() == 'not_found':
            raise ex.UnityResourceNotFoundError()
        elif self.get_id() == 'snap_in_use':
            raise ex.UnityDeleteAttachedSnapError()
        elif self.name == 'empty-host':
            raise ex.HostDeleteIsCalled()
        elif self.get_id() == 'lun_in_replication':
            if not force_snap_delete:
                raise ex.UnityDeleteLunInReplicationError()
        elif self.get_id() == 'lun_rep_session_1':
            raise ex.UnityResourceNotFoundError()

    @property
    def pool(self):
        return MockResource('pool0')

    @property
    def iscsi_host_initiators(self):
        iscsi_initiator = MockResource('iscsi_initiator')
        iscsi_initiator.initiator_id = ['iqn.1-1.com.e:c.host.0',
                                        'iqn.1-1.com.e:c.host.1']
        return iscsi_initiator

    @property
    def total_size_gb(self):
        return self.size_total / units.Gi

    @total_size_gb.setter
    def total_size_gb(self, value):
        if value == self.total_size_gb:
            raise ex.UnityNothingToModifyError()
        else:
            self.size_total = value * units.Gi

    def add_initiator(self, uid, force_create=None):
        self.initiator_id.append(uid)

    def attach(self, lun_or_snap, skip_hlu_0=True):
        if lun_or_snap.get_id() == 'already_attached':
            raise ex.UnityResourceAlreadyAttachedError()
        self.alu_hlu_map[lun_or_snap.get_id()] = len(self.alu_hlu_map)
        return self.get_hlu(lun_or_snap)

    @staticmethod
    def detach(lun_or_snap):
        if lun_or_snap.name == 'detach_failure':
            raise ex.DetachIsCalled()

    @staticmethod
    def detach_from(host):
        if host is None:
            raise ex.DetachFromIsCalled()

    def get_hlu(self, lun):
        return self.alu_hlu_map.get(lun.get_id(), None)

    @staticmethod
    def create_lun(lun_name, size_gb, description=None, io_limit_policy=None,
                   is_thin=None, is_compression=None, tiering_policy=None):
        if lun_name == 'in_use':
            raise ex.UnityLunNameInUseError()
        ret = MockResource(lun_name, 'lun_2')
        if io_limit_policy is not None:
            ret.max_iops = io_limit_policy.max_iops
            ret.max_kbps = io_limit_policy.max_kbps
        if is_thin is not None:
            ret.is_thin = is_thin
        if tiering_policy is not None:
            ret.tiering_policy = tiering_policy
        return ret

    @staticmethod
    def create_snap(name, is_auto_delete=False):
        if name == 'in_use':
            raise ex.UnitySnapNameInUseError()
        ret = MockResource(name)
        ret.is_auto_delete = is_auto_delete
        return ret

    @staticmethod
    def update(data=None):
        pass

    @property
    def iscsi_node(self):
        name = 'iqn.1-1.com.e:c.%s.0' % self.name
        return MockResource(name)

    @property
    def fc_host_initiators(self):
        init0 = MockResource('fhi_0')
        init0.initiator_id = '00:11:22:33:44:55:66:77:88:99:AA:BB:CC:CD:EE:FF'
        init1 = MockResource('fhi_1')
        init1.initiator_id = '00:11:22:33:44:55:66:77:88:99:AA:BB:BC:CD:EE:FF'
        return MockResourceList.create(init0, init1)

    @property
    def paths(self):
        path0 = MockResource('%s_path_0' % self.name)
        path0.is_logged_in = True
        path1 = MockResource('%s_path_1' % self.name)
        path1.is_logged_in = False
        path2 = MockResource('%s_path_2' % self.name)
        path2.is_logged_in = True
        return MockResourceList.create(path0, path1)

    @property
    def fc_port(self):
        ret = MockResource(_id='spa_iom_0_fc0')
        ret.wwn = '00:11:22:33:44:55:66:77:88:99:AA:BB:CC:DD:EE:FF'
        return ret

    @property
    def host_luns(self):
        if self.name == 'host-no-host_luns':
            return None
        return []

    @property
    def storage_resource(self):
        if self._storage_resource is None:
            self._storage_resource = MockResource(_id='sr_%s' % self._id,
                                                  name='sr_%s' % self.name)
        return self._storage_resource

    @storage_resource.setter
    def storage_resource(self, value):
        self._storage_resource = value

    def modify(self, name=None, is_compression=None, io_limit_policy=None):
        self.name = name

    def remove_from_storage(self, lun):
        pass

    def thin_clone(self, name, io_limit_policy=None, description=None):
        if name == 'thin_clone_name_in_use':
            raise ex.UnityLunNameInUseError
        return MockResource(_id=name, name=name)

    def get_snap(self, name):
        return MockResource(_id=name, name=name)

    def restore(self, delete_backup):
        return MockResource(_id='snap_1', name="internal_snap")

    def migrate(self, dest_pool, **kwargs):
        if dest_pool.id == 'fail_migration_pool':
            return False
        return True

    def replicate_cg_with_dst_resource_provisioning(self,
                                                    max_time_out_of_sync,
                                                    source_luns,
                                                    dst_pool_id,
                                                    remote_system=None,
                                                    dst_cg_name=None):
        return {'max_time_out_of_sync': max_time_out_of_sync,
                'dst_pool_id': dst_pool_id,
                'remote_system': remote_system,
                'dst_cg_name': dst_cg_name}

    def replicate_with_dst_resource_provisioning(self, max_time_out_of_sync,
                                                 dst_pool_id,
                                                 remote_system=None,
                                                 dst_lun_name=None):
        return {'max_time_out_of_sync': max_time_out_of_sync,
                'dst_pool_id': dst_pool_id,
                'remote_system': remote_system,
                'dst_lun_name': dst_lun_name}

    def failover(self, sync=None):
        return {'sync': sync}

    def failback(self, force_full_copy=None):
        return {'force_full_copy': force_full_copy}

    def check_cg_is_replicated(self):
        if self.name == 'replicated_cg':
            return True
        return False


class MockResourceList(object):
    def __init__(self, names=None, ids=None):
        if names is not None:
            self.resources = [MockResource(name=name) for name in names]
        elif ids is not None:
            self.resources = [MockResource(_id=_id) for _id in ids]

    @staticmethod
    def create(*rsc_list):
        ret = MockResourceList([])
        ret.resources = rsc_list
        return ret

    @property
    def name(self):
        return map(lambda i: i.name, self.resources)

    @property
    def list(self):
        return self.resources

    @list.setter
    def list(self, value):
        self.resources = []

    def __iter__(self):
        return self.resources.__iter__()

    def __len__(self):
        return len(self.resources)

    def __getattr__(self, item):
        return [getattr(i, item) for i in self.resources]

    def shadow_copy(self, **kwargs):
        if list(filter(None, kwargs.values())):
            return MockResourceList.create(self.resources[0])
        else:
            return self


class MockSystem(object):
    def __init__(self):
        self.serial_number = 'SYSTEM_SERIAL'
        self.system_version = '4.1.0'

    @property
    def info(self):
        mocked_info = mock.Mock()
        mocked_info.name = self.serial_number
        return mocked_info

    @staticmethod
    def get_lun(_id=None, name=None):
        if _id == 'not_found':
            raise ex.UnityResourceNotFoundError()
        if _id == 'tc_80':  # for thin clone with extending size
            lun = MockResource(name=_id, _id=_id)
            lun.total_size_gb = 7
            return lun
        return MockResource(name, _id)

    @staticmethod
    def get_pool(_id=None, name=None):
        if name == 'Pool 3':
            return MockResource(name, 'pool_3')
        if name or _id:
            return MockResource(name, _id)
        return MockResourceList(['Pool 1', 'Pool 2'])

    @staticmethod
    def get_snap(name):
        if name == 'not_found':
            raise ex.UnityResourceNotFoundError()
        return MockResource(name)

    @staticmethod
    def get_cg(name):
        if not name:
            raise ex.UnityResourceNotFoundError()
        return MockResource(name, _id=name)

    @staticmethod
    def create_host(name):
        return MockResource(name)

    @staticmethod
    def get_host(name):
        if name == 'not_found':
            raise ex.UnityResourceNotFoundError()
        if name == 'host1':
            ret = MockResource(name)
            ret.initiator_id = ['old-iqn']
            return ret
        return MockResource(name)

    @staticmethod
    def get_iscsi_portal():
        portal0 = MockResource('p0')
        portal0.ip_address = '1.1.1.1'
        portal1 = MockResource('p1')
        portal1.ip_address = '1.1.1.2'
        return MockResourceList.create(portal0, portal1)

    @staticmethod
    def get_fc_port():
        port0 = MockResource('fcp0')
        port0.wwn = '00:11:22:33:44:55:66:77:88:99:AA:BB:CC:DD:EE:FF'
        port1 = MockResource('fcp1')
        port1.wwn = '00:11:22:33:44:55:66:77:88:99:AA:BB:CC:DD:FF:EE'
        return MockResourceList.create(port0, port1)

    @staticmethod
    def create_io_limit_policy(name, max_iops=None, max_kbps=None):
        if name == 'in_use':
            raise ex.UnityPolicyNameInUseError()
        ret = MockResource(name)
        ret.max_iops = max_iops
        ret.max_kbps = max_kbps
        return ret

    @staticmethod
    def get_io_limit_policy(name):
        return MockResource(name=name)

    def get_remote_system(self, name=None):
        if name == 'not-exist':
            raise ex.UnityResourceNotFoundError()
        else:
            return {'name': name}

    def get_replication_session(self, name=None,
                                src_resource_id=None, dst_resource_id=None):
        if name == 'not-exist':
            raise ex.UnityResourceNotFoundError()
        elif src_resource_id == 'lun_in_replication':
            return [MockResource(name='rep_session')]
        elif src_resource_id == 'lun_not_in_replication':
            raise ex.UnityResourceNotFoundError()
        elif src_resource_id == 'lun_in_multiple_replications':
            return [MockResource(_id='lun_rep_session_1'),
                    MockResource(_id='lun_rep_session_2')]
        elif src_resource_id and ('is_in_replication'
                                  in src_resource_id):
            return [MockResource(name='rep_session')]
        elif dst_resource_id and ('is_in_replication'
                                  in dst_resource_id):
            return [MockResource(name='rep_session')]
        else:
            return {'name': name,
                    'src_resource_id': src_resource_id,
                    'dst_resource_id': dst_resource_id}


@mock.patch.object(client, 'storops', new='True')
def get_client():
    ret = client.UnityClient('1.2.3.4', 'user', 'pass')
    ret._system = MockSystem()
    return ret


########################
#
#   Start of Tests
#
########################
@ddt.ddt
@mock.patch.object(client, 'storops_ex', new=ex)
class ClientTest(unittest.TestCase):
    def setUp(self):
        self.client = get_client()

    def test_get_serial(self):
        self.assertEqual('SYSTEM_SERIAL', self.client.get_serial())

    def test_create_lun_success(self):
        name = 'LUN 3'
        pool = MockResource('Pool 0')
        lun = self.client.create_lun(name, 5, pool)
        self.assertEqual(name, lun.name)

    def test_create_lun_name_in_use(self):
        name = 'in_use'
        pool = MockResource('Pool 0')
        lun = self.client.create_lun(name, 6, pool)
        self.assertEqual('in_use', lun.name)

    def test_create_lun_with_io_limit(self):
        pool = MockResource('Pool 0')
        limit = MockResource('limit')
        limit.max_kbps = 100
        lun = self.client.create_lun('LUN 4', 6, pool, io_limit_policy=limit)
        self.assertEqual(100, lun.max_kbps)

    def test_create_lun_thick(self):
        name = 'thick_lun'
        pool = MockResource('Pool 0')
        lun = self.client.create_lun(name, 6, pool, is_thin=False)
        self.assertIsNotNone(lun.is_thin)
        self.assertFalse(lun.is_thin)
        self.assertIsNone(lun.tiering_policy)

    def test_create_auto_tier_lun(self):
        name = 'auto_tier_lun'
        tiering_policy = enums.TieringPolicyEnum.AUTOTIER
        pool = MockResource('Pool 0')
        lun = self.client.create_lun(name, 6, pool,
                                     tiering_policy=tiering_policy)
        self.assertIsNotNone(lun.tiering_policy)
        self.assertEqual(enums.TieringPolicyEnum.AUTOTIER, lun.tiering_policy)

    def test_create_high_tier_lun(self):
        name = 'high_tier_lun'
        tiering_policy = enums.TieringPolicyEnum.HIGHEST
        pool = MockResource('Pool 0')
        lun = self.client.create_lun(name, 6, pool,
                                     tiering_policy=tiering_policy)
        self.assertIsNotNone(lun.tiering_policy)
        self.assertEqual(enums.TieringPolicyEnum.HIGHEST, lun.tiering_policy)

    def test_thin_clone_success(self):
        name = 'tc_77'
        src_lun = MockResource(_id='id_77')
        lun = self.client.thin_clone(src_lun, name)
        self.assertEqual(name, lun.name)

    def test_thin_clone_name_in_used(self):
        name = 'thin_clone_name_in_use'
        src_lun = MockResource(_id='id_79')
        lun = self.client.thin_clone(src_lun, name)
        self.assertEqual(name, lun.name)

    def test_thin_clone_extend_size(self):
        name = 'tc_80'
        src_lun = MockResource(_id='id_80')
        lun = self.client.thin_clone(src_lun, name, io_limit_policy=None,
                                     new_size_gb=7)
        self.assertEqual(name, lun.name)
        self.assertEqual(7, lun.total_size_gb)

    def test_delete_lun_normal(self):
        self.assertIsNone(self.client.delete_lun('lun3'))

    def test_delete_lun_not_found(self):
        try:
            self.client.delete_lun('not_found')
        except ex.StoropsException:
            self.fail('not found error should be dealt with silently.')

    def test_delete_lun_in_replication(self):
        self.client.delete_lun('lun_in_replication')

    @ddt.data({'lun_id': 'lun_not_in_replication'},
              {'lun_id': 'lun_in_multiple_replications'})
    @ddt.unpack
    def test_delete_lun_replications(self, lun_id):
        self.client.delete_lun_replications(lun_id)

    def test_get_lun_with_id(self):
        lun = self.client.get_lun('lun4')
        self.assertEqual('lun4', lun.get_id())

    def test_get_lun_with_name(self):
        lun = self.client.get_lun(name='LUN 4')
        self.assertEqual('LUN 4', lun.name)

    def test_get_lun_not_found(self):
        ret = self.client.get_lun(lun_id='not_found')
        self.assertIsNone(ret)

    def test_get_pools(self):
        pools = self.client.get_pools()
        self.assertEqual(2, len(pools))

    def test_create_snap_normal(self):
        snap = self.client.create_snap('lun_1', 'snap_1')
        self.assertEqual('snap_1', snap.name)

    def test_create_snap_in_use(self):
        snap = self.client.create_snap('lun_1', 'in_use')
        self.assertEqual('in_use', snap.name)

    def test_delete_snap_error(self):
        def f():
            snap = MockResource(_id='snap_2')
            self.client.delete_snap(snap)

        self.assertRaises(ex.SnapDeleteIsCalled, f)

    def test_delete_snap_not_found(self):
        try:
            snap = MockResource(_id='not_found')
            self.client.delete_snap(snap)
        except ex.StoropsException:
            self.fail('snap not found should not raise exception.')

    def test_delete_snap_none(self):
        try:
            ret = self.client.delete_snap(None)
            self.assertIsNone(ret)
        except ex.StoropsException:
            self.fail('delete none should not raise exception.')

    def test_delete_snap_in_use(self):
        def f():
            snap = MockResource(_id='snap_in_use')
            self.client.delete_snap(snap)

        self.assertRaises(ex.UnityDeleteAttachedSnapError, f)

    def test_get_snap_found(self):
        snap = self.client.get_snap('snap_2')
        self.assertEqual('snap_2', snap.name)

    def test_get_snap_not_found(self):
        ret = self.client.get_snap('not_found')
        self.assertIsNone(ret)

    @mock.patch.object(coordination.Coordinator, 'get_lock')
    def test_create_host_found(self, fake_coordination):
        host = self.client.create_host('host1')

        self.assertEqual('host1', host.name)
        self.assertLessEqual(['iqn.1-1.com.e:c.a.a0'], host.initiator_id)

    @mock.patch.object(coordination.Coordinator, 'get_lock')
    def test_create_host_not_found(self, fake):
        host = self.client.create_host('not_found')
        self.assertEqual('not_found', host.name)
        self.assertIn('not_found', self.client.host_cache)

    def test_attach_lun(self):
        lun = MockResource(_id='lun1', name='l1')
        host = MockResource('host1')
        self.assertEqual(1, self.client.attach(host, lun))

    def test_attach_already_attached(self):
        lun = MockResource(_id='already_attached')
        host = MockResource('host1')
        hlu = self.client.attach(host, lun)
        self.assertEqual(99, hlu)

    def test_detach_lun(self):
        def f():
            lun = MockResource('detach_failure')
            host = MockResource('host1')
            self.client.detach(host, lun)

        self.assertRaises(ex.DetachIsCalled, f)

    def test_detach_all(self):
        def f():
            lun = MockResource('lun_44')
            self.client.detach_all(lun)

        self.assertRaises(ex.DetachFromIsCalled, f)

    @mock.patch.object(coordination.Coordinator, 'get_lock')
    def test_create_host(self, fake):
        self.assertEqual('host2', self.client.create_host('host2').name)

    @mock.patch.object(coordination.Coordinator, 'get_lock')
    def test_create_host_in_cache(self, fake):
        self.client.host_cache['already_in'] = MockResource(name='already_in')
        host = self.client.create_host('already_in')
        self.assertIn('already_in', self.client.host_cache)
        self.assertEqual('already_in', host.name)

    def test_update_host_initiators(self):
        host = MockResource(name='host_init')
        host = self.client.update_host_initiators(host, 'fake-iqn-1')

    def test_get_iscsi_target_info(self):
        ret = self.client.get_iscsi_target_info()
        expected = [{'iqn': 'iqn.1-1.com.e:c.p0.0', 'portal': '1.1.1.1:3260'},
                    {'iqn': 'iqn.1-1.com.e:c.p1.0', 'portal': '1.1.1.2:3260'}]
        self.assertListEqual(expected, ret)

    def test_get_iscsi_target_info_allowed_ports(self):
        ret = self.client.get_iscsi_target_info(allowed_ports=['spa_eth0'])
        expected = [{'iqn': 'iqn.1-1.com.e:c.p0.0', 'portal': '1.1.1.1:3260'}]
        self.assertListEqual(expected, ret)

    def test_get_fc_target_info_without_host(self):
        ret = self.client.get_fc_target_info()
        self.assertListEqual(['8899AABBCCDDEEFF', '8899AABBCCDDFFEE'],
                             sorted(ret))

    def test_get_fc_target_info_without_host_but_allowed_ports(self):
        ret = self.client.get_fc_target_info(allowed_ports=['spa_fc0'])
        self.assertListEqual(['8899AABBCCDDEEFF'], ret)

    def test_get_fc_target_info_with_host(self):
        host = MockResource('host0')
        ret = self.client.get_fc_target_info(host, True)
        self.assertListEqual(['8899AABBCCDDEEFF'], ret)

    def test_get_fc_target_info_with_host_and_allowed_ports(self):
        host = MockResource('host0')
        ret = self.client.get_fc_target_info(host, True,
                                             allowed_ports=['spb_iom_0_fc0'])
        self.assertListEqual([], ret)

    def test_get_io_limit_policy_none(self):
        ret = self.client.get_io_limit_policy(None)
        self.assertIsNone(ret)

    def test_get_io_limit_policy_create_new(self):
        specs = {'maxBWS': 2, 'id': 'max_2_mbps', 'maxIOPS': None}
        limit = self.client.get_io_limit_policy(specs)
        self.assertEqual('max_2_mbps', limit.name)
        self.assertEqual(2, limit.max_kbps)

    def test_create_io_limit_policy_success(self):
        limit = self.client.create_io_limit_policy('3kiops', max_iops=3000)
        self.assertEqual('3kiops', limit.name)
        self.assertEqual(3000, limit.max_iops)

    def test_create_io_limit_policy_in_use(self):
        limit = self.client.create_io_limit_policy('in_use', max_iops=100)
        self.assertEqual('in_use', limit.name)

    def test_expand_lun_success(self):
        lun = self.client.extend_lun('ev_3', 6)
        self.assertEqual(6, lun.total_size_gb)

    def test_expand_lun_nothing_to_modify(self):
        lun = self.client.extend_lun('ev_4', 5)
        self.assertEqual(5, lun.total_size_gb)

    def test_migrate_lun_success(self):
        ret = self.client.migrate_lun('lun_0', 'pool_1')
        self.assertTrue(ret)

    def test_migrate_lun_failed(self):
        ret = self.client.migrate_lun('lun_0', 'fail_migration_pool')
        self.assertFalse(ret)

    def test_migrate_lun_thick(self):
        ret = self.client.migrate_lun('lun_thick', 'pool_2', 'thick')
        self.assertTrue(ret)

    def test_migrate_lun_compressed(self):
        ret = self.client.migrate_lun('lun_compressed', 'pool_2', 'compressed')
        self.assertTrue(ret)

    def test_get_pool_id_by_name(self):
        self.assertEqual('pool_3', self.client.get_pool_id_by_name('Pool 3'))

    def test_get_pool_name(self):
        self.assertEqual('Pool0', self.client.get_pool_name('lun_0'))

    def test_restore_snapshot(self):
        back_snap = self.client.restore_snapshot('snap1')
        self.assertEqual("internal_snap", back_snap.name)

    def test_delete_host_wo_lock(self):
        host = MockResource(name='empty-host')
        self.client.host_cache['empty-host'] = host
        self.assertRaises(ex.HostDeleteIsCalled,
                          self.client.delete_host_wo_lock,
                          host)

    def test_delete_host_wo_lock_remove_from_cache(self):
        host = MockResource(name='empty-host-in-cache')
        self.client.host_cache['empty-host-in-cache'] = host
        self.client.delete_host_wo_lock(host)
        self.assertNotIn(host.name, self.client.host_cache)

    @ddt.data(('cg_1', 'cg_1_description', [MockResource(_id='sv_1')]),
              ('cg_2', None, None),
              ('cg_3', None, [MockResource(_id='sv_2')]),
              ('cg_4', 'cg_4_description', None))
    @ddt.unpack
    def test_create_cg(self, cg_name, cg_description, lun_add):
        created_cg = MockResource(_id='cg_1')
        with mock.patch.object(self.client.system, 'create_cg',
                               create=True, return_value=created_cg
                               ) as mocked_create:
            ret = self.client.create_cg(cg_name, description=cg_description,
                                        lun_add=lun_add)
            mocked_create.assert_called_once_with(cg_name,
                                                  description=cg_description,
                                                  lun_add=lun_add)
            self.assertEqual(created_cg, ret)

    def test_create_cg_existing_name(self):
        existing_cg = MockResource(_id='cg_1')
        with mock.patch.object(
                self.client.system, 'create_cg',
                side_effect=ex.UnityConsistencyGroupNameInUseError,
                create=True) as mocked_create, \
                mock.patch.object(self.client.system, 'get_cg',
                                  create=True,
                                  return_value=existing_cg) as mocked_get:
            ret = self.client.create_cg('existing_name')
            mocked_create.assert_called_once_with('existing_name',
                                                  description=None,
                                                  lun_add=None)
            mocked_get.assert_called_once_with(name='existing_name')
            self.assertEqual(existing_cg, ret)

    def test_get_cg(self):
        existing_cg = MockResource(_id='cg_1')
        with mock.patch.object(self.client.system, 'get_cg',
                               create=True,
                               return_value=existing_cg) as mocked_get:
            ret = self.client.get_cg('existing_name')
            mocked_get.assert_called_once_with(name='existing_name')
            self.assertEqual(existing_cg, ret)

    def test_get_cg_not_found(self):
        with mock.patch.object(self.client.system, 'get_cg',
                               create=True,
                               side_effect=ex.UnityResourceNotFoundError
                               ) as mocked_get:
            ret = self.client.get_cg('not_found_name')
            mocked_get.assert_called_once_with(name='not_found_name')
            self.assertIsNone(ret)

    def test_delete_cg(self):
        existing_cg = MockResource(_id='cg_1')
        with mock.patch.object(existing_cg, 'delete', create=True
                               ) as mocked_delete, \
                mock.patch.object(self.client, 'get_cg',
                                  create=True,
                                  return_value=existing_cg) as mocked_get:
            ret = self.client.delete_cg('cg_1_name')
            mocked_get.assert_called_once_with('cg_1_name')
            mocked_delete.assert_called_once()
            self.assertIsNone(ret)

    def test_update_cg(self):
        existing_cg = MockResource(_id='cg_1')
        lun_1 = MockResource(_id='sv_1')
        lun_2 = MockResource(_id='sv_2')
        lun_3 = MockResource(_id='sv_3')

        def _mocked_get_lun(lun_id):
            if lun_id == 'sv_1':
                return lun_1
            if lun_id == 'sv_2':
                return lun_2
            if lun_id == 'sv_3':
                return lun_3

        with mock.patch.object(existing_cg, 'update_lun', create=True
                               ) as mocked_update, \
                mock.patch.object(self.client, 'get_cg',
                                  create=True,
                                  return_value=existing_cg) as mocked_get, \
                mock.patch.object(self.client, 'get_lun',
                                  side_effect=_mocked_get_lun):
            ret = self.client.update_cg('cg_1_name', ['sv_1', 'sv_2'],
                                        ['sv_3'])
            mocked_get.assert_called_once_with('cg_1_name')
            mocked_update.assert_called_once_with(add_luns=[lun_1, lun_2],
                                                  remove_luns=[lun_3])
            self.assertIsNone(ret)

    def test_update_cg_empty_lun_ids(self):
        existing_cg = MockResource(_id='cg_1')
        with mock.patch.object(existing_cg, 'update_lun', create=True
                               ) as mocked_update, \
                mock.patch.object(self.client, 'get_cg',
                                  create=True,
                                  return_value=existing_cg) as mocked_get:
            ret = self.client.update_cg('cg_1_name', set(), set())
            mocked_get.assert_called_once_with('cg_1_name')
            mocked_update.assert_called_once_with(add_luns=[], remove_luns=[])
            self.assertIsNone(ret)

    def test_create_cg_group(self):
        existing_cg = MockResource(_id='cg_1')
        created_snap = MockResource(_id='snap_cg_1', name='snap_name_cg_1')
        with mock.patch.object(existing_cg, 'create_snap', create=True,
                               return_value=created_snap) as mocked_create, \
                mock.patch.object(self.client, 'get_cg',
                                  create=True,
                                  return_value=existing_cg) as mocked_get:
            ret = self.client.create_cg_snap('cg_1_name',
                                             snap_name='snap_name_cg_1')
            mocked_get.assert_called_once_with('cg_1_name')
            mocked_create.assert_called_once_with(name='snap_name_cg_1',
                                                  is_auto_delete=False)
            self.assertEqual(created_snap, ret)

    def test_create_cg_group_none_name(self):
        existing_cg = MockResource(_id='cg_1')
        created_snap = MockResource(_id='snap_cg_1')
        with mock.patch.object(existing_cg, 'create_snap', create=True,
                               return_value=created_snap) as mocked_create, \
                mock.patch.object(self.client, 'get_cg',
                                  create=True,
                                  return_value=existing_cg) as mocked_get:
            ret = self.client.create_cg_snap('cg_1_name')
            mocked_get.assert_called_once_with('cg_1_name')
            mocked_create.assert_called_once_with(name=None,
                                                  is_auto_delete=False)
            self.assertEqual(created_snap, ret)

    def test_filter_snaps_in_cg_snap(self):
        snaps = [MockResource(_id='snap_{}'.format(n)) for n in (1, 2)]
        snap_list = mock.MagicMock()
        snap_list.list = snaps
        with mock.patch.object(self.client.system, 'get_snap',
                               create=True,
                               return_value=snap_list) as mocked_get:
            ret = self.client.filter_snaps_in_cg_snap('snap_cg_1')
            mocked_get.assert_called_once_with(snap_group='snap_cg_1')
            self.assertEqual(snaps, ret)

    def test_create_replication(self):
        remote_system = MockResource(_id='RS_1')
        lun = MockResource(_id='sv_1')
        called = self.client.create_replication(lun, 60, 'pool_1',
                                                remote_system)
        self.assertEqual(called['max_time_out_of_sync'], 60)
        self.assertEqual(called['dst_pool_id'], 'pool_1')
        self.assertIs(called['remote_system'], remote_system)

    def test_get_remote_system(self):
        called = self.client.get_remote_system(name='remote-unity')
        self.assertEqual(called['name'], 'remote-unity')

    def test_get_remote_system_not_exist(self):
        called = self.client.get_remote_system(name='not-exist')
        self.assertIsNone(called)

    def test_get_replication_session(self):
        called = self.client.get_replication_session(name='rep-name')
        self.assertEqual(called['name'], 'rep-name')

    def test_get_replication_session_not_exist(self):
        self.assertRaises(client.ClientReplicationError,
                          self.client.get_replication_session,
                          name='not-exist')

    def test_failover_replication(self):
        rep_session = MockResource(_id='rep_id_1')
        called = self.client.failover_replication(rep_session)
        self.assertFalse(called['sync'])

    def test_failover_replication_raise(self):
        rep_session = MockResource(_id='rep_id_1')

        def mock_failover(sync=None):
            raise ex.UnityResourceNotFoundError()

        rep_session.failover = mock_failover
        self.assertRaises(client.ClientReplicationError,
                          self.client.failover_replication,
                          rep_session)

    def test_failback_replication(self):
        rep_session = MockResource(_id='rep_id_1')
        called = self.client.failback_replication(rep_session)
        self.assertTrue(called['force_full_copy'])

    def test_failback_replication_raise(self):
        rep_session = MockResource(_id='rep_id_1')

        def mock_failback(force_full_copy=None):
            raise ex.UnityResourceNotFoundError()

        rep_session.failback = mock_failback
        self.assertRaises(client.ClientReplicationError,
                          self.client.failback_replication,
                          rep_session)

    def test_create_cg_replication(self):
        remote_system = MockResource(_id='RS_2')
        cg_name = 'test_cg'
        called = self.client.create_cg_replication(
            cg_name, 'pool_1', remote_system, 60)
        self.assertEqual(60, called['max_time_out_of_sync'])
        self.assertEqual('pool_1', called['dst_pool_id'])
        self.assertEqual('test_cg', called['dst_cg_name'])
        self.assertIs(remote_system, called['remote_system'])

    def test_cg_in_replciation(self):
        existing_cg = MockResource(_id='replicated_cg')
        result = self.client.is_cg_replicated(existing_cg.id)
        self.assertTrue(result)

    def test_cg_not_in_replciation(self):
        existing_cg = MockResource(_id='test_cg')
        result = self.client.is_cg_replicated(existing_cg.id)
        self.assertFalse(result)

    def test_delete_cg_rep_session(self):
        src_cg = MockResource(_id='cg_is_in_replication')
        result = self.client.delete_cg_rep_session(src_cg.id)
        self.assertIsNone(result)

    def test_failover_cg_rep_session(self):
        src_cg = MockResource(_id='failover_cg_is_in_replication')
        result = self.client.failover_cg_rep_session(src_cg.id, True)
        self.assertIsNone(result)

    def test_failback_cg_rep_session(self):
        src_cg = MockResource(_id='failback_cg_is_in_replication')
        result = self.client.failback_cg_rep_session(src_cg.id)
        self.assertIsNone(result)



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt

from cinder import exception
from cinder.tests.unit.volume.drivers.dell_emc import powerflex
from cinder.volume import configuration


@ddt.ddt
class TestReplication(powerflex.TestPowerFlexDriver):
    """Test cases for PowerFlex replication support."""

    def setUp(self):
        super(TestReplication, self).setUp()

        self.replication_backend_id = 'powerflex_repl'
        replication_device = [
            {
                'backend_id': self.replication_backend_id,
                'san_ip': '127.0.0.2',
                'san_login': 'test',
                'san_password': 'pass'
            }
        ]
        self.override_config('replication_device',
                             override=replication_device,
                             group=configuration.SHARED_CONF_GROUP)

        self.HTTPS_MOCK_RESPONSES = {
            self.RESPONSE_MODE.Valid: {
                'types/Domain/instances/getByName::' + self.PROT_DOMAIN_NAME:
                    '"{}"'.format(self.PROT_DOMAIN_ID),
                'types/Pool/instances/getByName::{},{}'.format(
                    self.PROT_DOMAIN_ID, self.STORAGE_POOL_NAME):
                    '"{}"'.format(self.STORAGE_POOL_ID),
                'instances/ProtectionDomain::{}'.format(self.PROT_DOMAIN_ID):
                    {'id': self.PROT_DOMAIN_ID},
                'instances/StoragePool::{}'.format(self.STORAGE_POOL_ID):
                    {'id': self.STORAGE_POOL_ID, 'zeroPaddingEnabled': True},
            },
        }

    def test_do_setup_replication_configured(self):
        super(powerflex.mocks.PowerFlexDriver, self.driver).do_setup({})
        self.driver.check_for_setup_error()
        self.assertTrue(self.driver.secondary_client.is_configured)
        self.assertTrue(self.driver.replication_enabled)

    @ddt.data(
        [
            {
                'backend_id': 'powerflex_repl1'
            },
            {
                'backend_id': 'powerflex_repl2'
            }
        ],
        [
            {
                'backend_id': 'powerflex_repl1',
                'san_ip': '127.0.0.2'
            },
        ]
    )
    def test_do_setup_replication_bad_configuration(self, replication_device):
        self.override_config('replication_device',
                             override=replication_device,
                             group=configuration.SHARED_CONF_GROUP)
        self.assertRaises(exception.InvalidInput,
                          super(powerflex.mocks.PowerFlexDriver,
                                self.driver).do_setup,
                          {})

    def test_do_setup_already_failed_over(self):
        self.driver.active_backend_id = 'powerflex_repl'
        super(powerflex.mocks.PowerFlexDriver, self.driver).do_setup({})
        self.driver.check_for_setup_error()
        self.assertFalse(self.driver.replication_enabled)

    def test_failover_host(self):
        self.test_do_setup_replication_configured()
        self.driver.failover_host({}, [], self.replication_backend_id)
        self.assertEqual(self.replication_backend_id,
                         self.driver.active_backend_id)

    def test_failover_host_failback(self):
        self.test_do_setup_already_failed_over()
        self.driver.failover_host({}, [], 'default')
        self.assertEqual('default', self.driver.active_backend_id)

    @ddt.data("not_valid_target", None)
    def test_failover_host_secondary_id_invalid(self, secondary_id):
        self.test_do_setup_replication_configured()
        self.assertRaises(exception.InvalidReplicationTarget,
                          self.driver.failover_host,
                          context={},
                          volumes=[],
                          secondary_id=secondary_id)



# Copyright (c) 2017 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from cinder.policies import base


# MANAGE_POLICY is deprecated
MANAGE_POLICY = 'group:group_types_manage'
CREATE_POLICY = 'group:group_types:create'
UPDATE_POLICY = 'group:group_types:update'
DELETE_POLICY = 'group:group_types:delete'
SHOW_ACCESS_POLICY = 'group:access_group_types_specs'
# SPEC_POLICY is deprecated
SPEC_POLICY = 'group:group_types_specs'
SPEC_GET_POLICY = 'group:group_types_specs:get'
SPEC_GET_ALL_POLICY = 'group:group_types_specs:get_all'
SPEC_CREATE_POLICY = 'group:group_types_specs:create'
SPEC_UPDATE_POLICY = 'group:group_types_specs:update'
SPEC_DELETE_POLICY = 'group:group_types_specs:delete'

deprecated_manage_policy = base.CinderDeprecatedRule(
    name=MANAGE_POLICY,
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=(f'{MANAGE_POLICY} has been replaced by more granular '
                       'policies that separately govern POST, PUT, and DELETE '
                       'operations.'),
)
deprecated_spec_policy = base.CinderDeprecatedRule(
    name=SPEC_POLICY,
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=(f'{SPEC_POLICY} has been replaced by more granular '
                       'policies that separately govern GET, POST, PUT, and '
                       'DELETE operations.'),
)

group_types_policies = [
    policy.DocumentedRuleDefault(
        name=CREATE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Create a group type.",
        operations=[
            {
                'method': 'POST',
                'path': '/group_types/'
            },
        ],
        deprecated_rule=deprecated_manage_policy,
    ),
    policy.DocumentedRuleDefault(
        name=UPDATE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Update a group type.",
        operations=[
            {
                'method': 'PUT',
                'path': '/group_types/{group_type_id}'
            },
        ],
        deprecated_rule=deprecated_manage_policy,
    ),
    policy.DocumentedRuleDefault(
        name=DELETE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Delete a group type.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/group_types/{group_type_id}'
            },
        ],
        deprecated_rule=deprecated_manage_policy,
    ),
    policy.DocumentedRuleDefault(
        name=SHOW_ACCESS_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Show group type with type specs attributes.",
        operations=[
            {
                'method': 'GET',
                'path': '/group_types/{group_type_id}'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=SPEC_GET_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Show a group type spec.",
        operations=[
            {
                'method': 'GET',
                'path': '/group_types/{group_type_id}/group_specs/{g_spec_id}'
            },
        ],
        deprecated_rule=deprecated_spec_policy,
    ),
    policy.DocumentedRuleDefault(
        name=SPEC_GET_ALL_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="List group type specs.",
        operations=[
            {
                'method': 'GET',
                'path': '/group_types/{group_type_id}/group_specs'
            },
        ],
        deprecated_rule=deprecated_spec_policy,
    ),
    policy.DocumentedRuleDefault(
        name=SPEC_CREATE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Create a group type spec.",
        operations=[
            {
                'method': 'POST',
                'path': '/group_types/{group_type_id}/group_specs'
            },
        ],
        deprecated_rule=deprecated_spec_policy,
    ),
    policy.DocumentedRuleDefault(
        name=SPEC_UPDATE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Update a group type spec.",
        operations=[
            {
                'method': 'PUT',
                'path': '/group_types/{group_type_id}/group_specs/{g_spec_id}'
            },
        ],
        deprecated_rule=deprecated_spec_policy,
    ),
    policy.DocumentedRuleDefault(
        name=SPEC_DELETE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Delete a group type spec.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/group_types/{group_type_id}/group_specs/{g_spec_id}'
            },
        ],
        deprecated_rule=deprecated_spec_policy,
    ),
]


def list_rules():
    return group_types_policies



# Copyright 2015 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""
Volume driver for IBM FlashSystem storage systems with FC protocol.

Limitations:
1. Cinder driver only works when open_access_enabled=off.

"""

import random
import threading

from oslo_log import log as logging
from oslo_utils import excutils
import six

from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder import utils
from cinder.volume.drivers.ibm import flashsystem_common as fscommon
from cinder.volume.drivers.san import san
from cinder.zonemanager import utils as fczm_utils

LOG = logging.getLogger(__name__)


@interface.volumedriver
class FlashSystemFCDriver(fscommon.FlashSystemDriver):
    """IBM FlashSystem FC volume driver.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver
        1.0.1 - Code clean up
        1.0.2 - Add lock into vdisk map/unmap, connection
                initialize/terminate
        1.0.3 - Initial driver for iSCSI
        1.0.4 - Split Flashsystem driver into common and FC
        1.0.5 - Report capability of volume multiattach
        1.0.6 - Fix bug #1469581, add I/T mapping check in
                terminate_connection
        1.0.7 - Fix bug #1505477, add host name check in
                _find_host_exhaustive for FC
        1.0.8 - Fix bug #1572743, multi-attach attribute
                should not be hardcoded, only in iSCSI
        1.0.9 - Fix bug #1570574, Cleanup host resource
                leaking, changes only in iSCSI
        1.0.10 - Fix bug #1585085, add host name check in
                 _find_host_exhaustive for iSCSI
        1.0.11 - Update driver to use ABC metaclasses
        1.0.12 - Update driver to support Manage/Unmanage
                 existing volume
    """

    VERSION = "1.0.12"

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "IBM_STORAGE_CI"

    def __init__(self, *args, **kwargs):
        super(FlashSystemFCDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(fscommon.flashsystem_opts)
        self.configuration.append_config_values(san.san_opts)

    def _check_vdisk_params(self, params):
        # Check that the requested protocol is enabled
        if params['protocol'] != self._protocol:
            msg = (_("Illegal value '%(prot)s' specified for "
                     "flashsystem_connection_protocol: "
                     "valid value(s) are %(enabled)s.")
                   % {'prot': params['protocol'],
                      'enabled': self._protocol})
            raise exception.InvalidInput(reason=msg)

    def _create_host(self, connector):
        """Create a new host on the storage system.

        We create a host and associate it with the given connection
        information.

        """

        LOG.debug('enter: _create_host: host %s.', connector['host'])

        rand_id = six.text_type(random.randint(0, 99999999)).zfill(8)
        host_name = '%s-%s' % (self._connector_to_hostname_prefix(connector),
                               rand_id)

        ports = []
        if 'FC' == self._protocol and 'wwpns' in connector:
            for wwpn in connector['wwpns']:
                ports.append('-hbawwpn %s' % wwpn)

        self._driver_assert(ports,
                            (_('_create_host: No connector ports.')))
        port1 = ports.pop(0)
        arg_name, arg_val = port1.split()
        ssh_cmd = ['svctask', 'mkhost', '-force', arg_name, arg_val, '-name',
                   '"%s"' % host_name]
        out, err = self._ssh(ssh_cmd)
        self._assert_ssh_return('successfully created' in out,
                                '_create_host', ssh_cmd, out, err)

        for port in ports:
            arg_name, arg_val = port.split()
            ssh_cmd = ['svctask', 'addhostport', '-force',
                       arg_name, arg_val, host_name]
            out, err = self._ssh(ssh_cmd)
            self._assert_ssh_return(
                (not out.strip()),
                '_create_host', ssh_cmd, out, err)

        LOG.debug(
            'leave: _create_host: host %(host)s - %(host_name)s.',
            {'host': connector['host'], 'host_name': host_name})

        return host_name

    def _find_host_exhaustive(self, connector, hosts):
        hname = connector['host']
        hnames = [ihost[0:ihost.rfind('-')] for ihost in hosts]
        if hname in hnames:
            host = hosts[hnames.index(hname)]
            ssh_cmd = ['svcinfo', 'lshost', '-delim', '!', host]
            out, err = self._ssh(ssh_cmd)
            self._assert_ssh_return(
                out.strip(),
                '_find_host_exhaustive', ssh_cmd, out, err)
            attr_lines = [attr_line for attr_line in out.split('\n')]
            attr_parm = {}
            for attr_line in attr_lines:
                attr_name, foo, attr_val = attr_line.partition('!')
                attr_parm[attr_name] = attr_val
            if ('WWPN' in attr_parm.keys() and 'wwpns' in connector and
                    attr_parm['WWPN'].lower() in
                    map(str.lower, map(str, connector['wwpns']))):
                return host
        else:
            LOG.warning('Host %(host)s was not found on backend storage.',
                        {'host': hname})
        return None

    def _get_conn_fc_wwpns(self):
        wwpns = []

        cmd = ['svcinfo', 'lsportfc']

        generator = self._port_conf_generator(cmd)
        header = next(generator, None)
        if not header:
            return wwpns

        for port_data in generator:
            try:
                if port_data['status'] == 'active':
                    wwpns.append(port_data['WWPN'])
            except KeyError:
                self._handle_keyerror('lsportfc', header)

        return wwpns

    def _get_fc_wwpns(self):
        for key in self._storage_nodes:
            node = self._storage_nodes[key]
            ssh_cmd = ['svcinfo', 'lsnode', '-delim', '!', node['id']]
            attributes = self._execute_command_and_parse_attributes(ssh_cmd)
            wwpns = set(node['WWPN'])
            for i, s in zip(attributes['port_id'], attributes['port_status']):
                if 'unconfigured' != s:
                    wwpns.add(i)
            node['WWPN'] = list(wwpns)
            LOG.info('WWPN on node %(node)s: %(wwpn)s.',
                     {'node': node['id'], 'wwpn': node['WWPN']})

    def _get_vdisk_map_properties(
            self, connector, lun_id, vdisk_name, vdisk_id, vdisk_params):
        """Get the map properties of vdisk."""

        LOG.debug(
            'enter: _get_vdisk_map_properties: vdisk '
            '%(vdisk_name)s.', {'vdisk_name': vdisk_name})

        IO_group = '0'

        io_group_nodes = []
        for k, node in self._storage_nodes.items():
            if vdisk_params['protocol'] != node['protocol']:
                continue
            if node['IO_group'] == IO_group:
                io_group_nodes.append(node)

        if not io_group_nodes:
            msg = (_('_get_vdisk_map_properties: No node found in '
                     'I/O group %(gid)s for volume %(vol)s.')
                   % {'gid': IO_group, 'vol': vdisk_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        properties = {}
        properties['target_discovered'] = False
        properties['target_lun'] = lun_id
        properties['volume_id'] = vdisk_id

        type_str = 'fibre_channel'
        conn_wwpns = self._get_conn_fc_wwpns()

        if not conn_wwpns:
            msg = _('_get_vdisk_map_properties: Could not get FC '
                    'connection information for the host-volume '
                    'connection. Is the host configured properly '
                    'for FC connections?')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        properties['target_wwn'] = conn_wwpns

        if "zvm_fcp" in connector:
            properties['zvm_fcp'] = connector['zvm_fcp']

        properties['initiator_target_map'] = self._build_initiator_target_map(
            connector['wwpns'], conn_wwpns)

        LOG.debug(
            'leave: _get_vdisk_map_properties: vdisk '
            '%(vdisk_name)s.', {'vdisk_name': vdisk_name})

        return {'driver_volume_type': type_str, 'data': properties}

    @utils.synchronized('flashsystem-init-conn', external=True)
    def initialize_connection(self, volume, connector):
        """Perform work so that an FC connection can be made.

        To be able to create a FC connection from a given host to a
        volume, we must:
        1. Translate the given WWNN to a host name
        2. Create new host on the storage system if it does not yet exist
        3. Map the volume to the host if it is not already done
        4. Return the connection information for relevant nodes (in the
           proper I/O group)

        """

        LOG.debug(
            'enter: initialize_connection: volume %(vol)s with '
            'connector %(conn)s.', {'vol': volume, 'conn': connector})

        vdisk_name = volume['name']
        vdisk_id = volume['id']
        vdisk_params = self._get_vdisk_params(volume['volume_type_id'])

        # TODO(edwin): might fix it after vdisk copy function is
        # ready in FlashSystem thin-provision layer. As this validation
        # is to check the vdisk which is in copying, at present in firmware
        # level vdisk doesn't allow to map host which it is copy. New
        # vdisk clone and snapshot function will cover it. After that the
        # _wait_vdisk_copy_completed need some modification.
        self._wait_vdisk_copy_completed(vdisk_name)

        self._driver_assert(
            self._is_vdisk_defined(vdisk_name),
            (_('initialize_connection: vdisk %s is not defined.')
             % vdisk_name))

        lun_id = self._map_vdisk_to_host(vdisk_name, connector)

        properties = {}
        try:
            properties = self._get_vdisk_map_properties(
                connector, lun_id, vdisk_name, vdisk_id, vdisk_params)
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                self.terminate_connection(volume, connector)
                LOG.error('initialize_connection: Failed to collect '
                          'return properties for volume %(vol)s and '
                          'connector %(conn)s.',
                          {'vol': volume, 'conn': connector})

        LOG.debug(
            'leave: initialize_connection:\n volume: %(vol)s\n connector '
            '%(conn)s\n properties: %(prop)s.',
            {'vol': volume,
             'conn': connector,
             'prop': properties})

        fczm_utils.add_fc_zone(properties)
        return properties

    @utils.synchronized('flashsystem-term-conn', external=True)
    def terminate_connection(self, volume, connector, **kwargs):
        """Cleanup after connection has been terminated.

        When we clean up a terminated connection between a given connector
        and volume, we:
        1. Translate the given connector to a host name
        2. Remove the volume-to-host mapping if it exists
        3. Delete the host if it has no more mappings (hosts are created
           automatically by this driver when mappings are created)
        """
        LOG.debug(
            'enter: terminate_connection: volume %(vol)s with '
            'connector %(conn)s.',
            {'vol': volume, 'conn': connector})

        return_data = {
            'driver_volume_type': 'fibre_channel',
            'data': {},
        }

        vdisk_name = volume['name']
        self._wait_vdisk_copy_completed(vdisk_name)
        self._unmap_vdisk_from_host(vdisk_name, connector)

        host_name = self._get_host_from_connector(connector)
        if not host_name:
            properties = {}
            conn_wwpns = self._get_conn_fc_wwpns()
            properties['target_wwn'] = conn_wwpns
            properties['initiator_target_map'] = (
                self._build_initiator_target_map(
                    connector['wwpns'], conn_wwpns))
            return_data['data'] = properties
            fczm_utils.remove_fc_zone(return_data)

        LOG.debug(
            'leave: terminate_connection: volume %(vol)s with '
            'connector %(conn)s.', {'vol': volume, 'conn': connector})

        return return_data

    def do_setup(self, ctxt):
        """Check that we have all configuration details from the storage."""

        self._context = ctxt

        # Get data of configured node
        self._get_node_data()

        # Get the WWPNs of the FlashSystem nodes
        self._get_fc_wwpns()

        # For each node, check what connection modes it supports.  Delete any
        # nodes that do not support any types (may be partially configured).
        to_delete = []
        for k, node in self._storage_nodes.items():
            if not node['WWPN']:
                to_delete.append(k)

        for delkey in to_delete:
            del self._storage_nodes[delkey]

        # Make sure we have at least one node configured
        self._driver_assert(self._storage_nodes,
                            'do_setup: No configured nodes.')

        self._protocol = node['protocol'] = 'FC'

        # Set for vdisk synchronization
        self._vdisk_copy_in_progress = set()
        self._vdisk_copy_lock = threading.Lock()
        self._check_lock_interval = 5

    def validate_connector(self, connector):
        """Check connector."""
        if 'FC' == self._protocol and 'wwpns' not in connector:
            LOG.error('The connector does not contain the '
                      'required information: wwpns is missing')
            raise exception.InvalidConnectorException(missing='wwpns')



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt

from cinder.api.contrib import admin_actions
from cinder.api.contrib import snapshot_actions
from cinder.api import microversions as mv
from cinder import exception
from cinder.policies import snapshot_actions as policy
from cinder.policies import snapshots as snapshots_policy
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit.policies import base
from cinder.tests.unit import utils as test_utils


@ddt.ddt
class SnapshotActionsPolicyTest(base.BasePolicyTest):
    authorized_members = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]

    unauthorized_members = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_admins = [
        'legacy_admin',
        'system_admin',
        'project_admin',
    ]

    unauthorized_admins = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_member',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    # DB validations will throw SnapshotNotFound for some contexts
    unauthorized_exceptions = [
        exception.SnapshotNotFound,
    ]

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = snapshot_actions.SnapshotActionsController()
        self.admin_controller = admin_actions.SnapshotAdminController()
        self.api_path = '/v3/%s/snapshots' % (self.project_id)
        self.api_version = mv.BASE_VERSION
        # Relax the snapshots GET_POLICY in order to get past that check.
        self.policy.set_rules({snapshots_policy.GET_POLICY: ""},
                              overwrite=False)

    def _create_snapshot(self, **kwargs):
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 name='fake_vol_type',
                                                 testcase_instance=self)

        volume = test_utils.create_volume(self.project_member_context,
                                          volume_type_id=vol_type.id,
                                          testcase_instance=self)

        snapshot = test_utils.create_snapshot(self.project_member_context,
                                              volume_id=volume.id,
                                              testcase_instance=self, **kwargs)
        return snapshot

    @ddt.data(*base.all_users)
    def test_reset_status_policy(self, user_id):
        snapshot = self._create_snapshot(status='error')
        rule_name = policy.RESET_STATUS_POLICY
        url = '%s/%s/action' % (self.api_path, snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-reset_status": {
                "status": "available",
            }
        }

        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.admin_controller._reset_status, req,
                                 id=snapshot.id, body=body)

    @ddt.data(*base.all_users)
    def test_update_status_policy(self, user_id):
        snapshot = self._create_snapshot(status='creating')
        rule_name = policy.UPDATE_STATUS_POLICY
        url = '%s/%s/action' % (self.api_path, snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-update_snapshot_status": {
                "status": "error"
            }
        }

        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.controller._update_snapshot_status, req,
                                 id=snapshot.id, body=body)

    @ddt.data(*base.all_users)
    def test_force_delete_policy(self, user_id):
        snapshot = self._create_snapshot(status='error')
        rule_name = policy.FORCE_DELETE_POLICY
        url = '%s/%s/action' % (self.api_path, snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-force_delete": {}
        }

        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.admin_controller._force_delete, req,
                                 id=snapshot.id, body=body)


class SnapshotActionsPolicySecureRbacTest(SnapshotActionsPolicyTest):
    authorized_members = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
    ]

    unauthorized_members = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)
